class TransactionSubmissionError(Exception):
    pass


class TransactionSubmissionPending(TransactionSubmissionError):
    pass


class TransactionSubmissionBlocked(TransactionSubmissionError):
    pass


class TransactionSubmissionFailed(TransactionSubmissionError):
    pass



"""
This module implements the logic for the authentication endpoint, as per SEP 10.
This defines a standard way for wallets and anchors to create authenticated web sessions
on behalf of a user who holds a Stellar account.

See: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0010.md
"""
import os
import jwt
import toml
from urllib.parse import urlparse

from django.utils.translation import gettext
from rest_framework import status
from rest_framework.views import APIView
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.parsers import JSONParser, MultiPartParser, FormParser
from rest_framework.renderers import JSONRenderer, BrowsableAPIRenderer
from stellar_sdk.operation import ManageData
from stellar_sdk.sep.stellar_toml import fetch_stellar_toml
from stellar_sdk.sep.stellar_web_authentication import (
    build_challenge_transaction,
    read_challenge_transaction,
    verify_challenge_transaction_threshold,
    verify_challenge_transaction_signed_by_client_master_key,
)
from stellar_sdk.sep.exceptions import (
    InvalidSep10ChallengeError,
    StellarTomlNotFoundError,
)
from stellar_sdk.exceptions import (
    NotFoundError,
    ConnectionError,
    Ed25519PublicKeyInvalidError,
)
from stellar_sdk import Keypair, MuxedAccount
from stellar_sdk.client.requests_client import RequestsClient

from polaris import settings
from polaris.utils import getLogger, render_error_response

MIME_URLENCODE, MIME_JSON = "application/x-www-form-urlencoded", "application/json"
logger = getLogger(__name__)


class SEP10Auth(APIView):
    """
    `GET /auth` can be used to get a challenge Stellar transaction.
    The client can then sign it using their private key and hit `POST /auth`
    to receive a JSON web token. That token can be used to authenticate calls
    to the other SEP 24 endpoints.
    """

    parser_classes = [JSONParser, MultiPartParser, FormParser]
    renderer_classes = [JSONRenderer, BrowsableAPIRenderer]

    ###############
    # GET functions
    ###############
    def get(self, request, *_args, **_kwargs) -> Response:
        account = request.GET.get("account")
        if not account:
            return Response(
                {"error": "no 'account' provided"}, status=status.HTTP_400_BAD_REQUEST
            )

        memo = request.GET.get("memo")
        if memo:
            try:
                memo = int(memo)
            except ValueError:
                return Response(
                    {"error": "invalid 'memo' value. Expected a 64-bit integer."},
                    status=status.HTTP_400_BAD_REQUEST,
                )
            if account.startswith("M"):
                return Response(
                    {
                        "error": "'memo' cannot be passed with a muxed client account address (M...)"
                    },
                    status=status.HTTP_400_BAD_REQUEST,
                )

        home_domain = request.GET.get("home_domain")
        if home_domain and home_domain not in settings.SEP10_HOME_DOMAINS:
            return Response(
                {
                    "error": f"invalid 'home_domain' value. Accepted values: {settings.SEP10_HOME_DOMAINS}"
                },
                status=status.HTTP_400_BAD_REQUEST,
            )
        elif not home_domain:
            home_domain = settings.SEP10_HOME_DOMAINS[0]

        client_domain, client_signing_key = request.GET.get("client_domain"), None
        if settings.SEP10_CLIENT_ATTRIBUTION_REQUIRED and not client_domain:
            return render_error_response(
                gettext("'client_domain' is required"), status_code=400
            )
        elif client_domain:
            if urlparse(f"https://{client_domain}").netloc != client_domain:
                return render_error_response(
                    gettext("'client_domain' must be a valid hostname"), status_code=400
                )
            elif (
                settings.SEP10_CLIENT_ATTRIBUTION_DENYLIST
                and client_domain in settings.SEP10_CLIENT_ATTRIBUTION_DENYLIST
            ) or (
                settings.SEP10_CLIENT_ATTRIBUTION_ALLOWLIST
                and client_domain not in settings.SEP10_CLIENT_ATTRIBUTION_ALLOWLIST
            ):
                if settings.SEP10_CLIENT_ATTRIBUTION_REQUIRED:
                    return render_error_response(
                        gettext("unrecognized 'client_domain'"), status_code=403
                    )
                else:
                    client_domain = None

        if client_domain:
            try:
                client_signing_key = self._get_client_signing_key(client_domain)
            except (
                ConnectionError,
                StellarTomlNotFoundError,
                toml.decoder.TomlDecodeError,
            ):
                return render_error_response(
                    gettext("unable to fetch 'client_domain' SIGNING_KEY"),
                )
            except ValueError as e:
                return render_error_response(str(e))

        try:
            transaction = self._challenge_transaction(
                account, home_domain, client_domain, client_signing_key, memo
            )
        except ValueError as e:
            return Response({"error": str(e)}, status=status.HTTP_400_BAD_REQUEST)

        logger.info(f"Returning SEP-10 challenge for account {account}")
        return Response(
            {
                "transaction": transaction,
                "network_passphrase": settings.STELLAR_NETWORK_PASSPHRASE,
            }
        )

    @staticmethod
    def _challenge_transaction(
        client_account,
        home_domain,
        client_domain=None,
        client_signing_key=None,
        memo=None,
    ):
        """
        Generate the challenge transaction for a client account.
        This is used in `GET <auth>`, as per SEP 10.
        Returns the XDR encoding of that transaction.
        """
        return build_challenge_transaction(
            server_secret=settings.SIGNING_SEED,
            client_account_id=client_account,
            home_domain=home_domain,
            web_auth_domain=urlparse(settings.HOST_URL).netloc,
            network_passphrase=settings.STELLAR_NETWORK_PASSPHRASE,
            timeout=900,
            client_domain=client_domain,
            client_signing_key=client_signing_key,
            memo=memo,
        )

    ################
    # POST functions
    ################
    def post(self, request: Request, *_args, **_kwargs) -> Response:
        envelope_xdr = request.data.get("transaction")
        if not envelope_xdr:
            return render_error_response(gettext("'transaction' is required"))
        client_domain, error_response = self._validate_challenge_xdr(envelope_xdr)
        if error_response:
            return error_response
        else:
            return Response({"token": self._generate_jwt(envelope_xdr, client_domain)})

    @staticmethod
    def _validate_challenge_xdr(envelope_xdr: str):
        """
        Validate the provided TransactionEnvelope XDR (base64 string).

        If the source account of the challenge transaction exists, verify the weight
        of the signers on the challenge are signers for the account and the medium
        threshold on the account is met by those signers.

        If the source account does not exist, verify that the keypair used as the
        source for the challenge transaction has signed the challenge. This is
        sufficient because newly created accounts have their own keypair as signer
        with a weight greater than the default thresholds.
        """
        logger.info("Validating challenge transaction")
        generic_err_msg = gettext("error while validating challenge: %s")
        try:
            challenge = read_challenge_transaction(
                challenge_transaction=envelope_xdr,
                server_account_id=settings.SIGNING_KEY,
                home_domains=settings.SEP10_HOME_DOMAINS,
                web_auth_domain=urlparse(settings.HOST_URL).netloc,
                network_passphrase=settings.STELLAR_NETWORK_PASSPHRASE,
            )
        except (InvalidSep10ChallengeError, TypeError) as e:
            return None, render_error_response(generic_err_msg % (str(e)))

        client_domain = None
        for operation in challenge.transaction.transaction.operations:
            if (
                isinstance(operation, ManageData)
                and operation.data_name == "client_domain"
            ):
                client_domain = operation.data_value.decode()
                break

        # extract the Stellar account from the muxed account to check for its existence
        stellar_account = challenge.client_account_id
        if challenge.client_account_id.startswith("M"):
            stellar_account = MuxedAccount.from_account(
                challenge.client_account_id
            ).account_id

        try:
            account = settings.HORIZON_SERVER.load_account(stellar_account)
        except NotFoundError:
            logger.info("Account does not exist, using client's master key to verify")
            try:
                verify_challenge_transaction_signed_by_client_master_key(
                    challenge_transaction=envelope_xdr,
                    server_account_id=settings.SIGNING_KEY,
                    home_domains=settings.SEP10_HOME_DOMAINS,
                    web_auth_domain=urlparse(settings.HOST_URL).netloc,
                    network_passphrase=settings.STELLAR_NETWORK_PASSPHRASE,
                )
                if (client_domain and len(challenge.transaction.signatures) != 3) or (
                    not client_domain and len(challenge.transaction.signatures) != 2
                ):
                    raise InvalidSep10ChallengeError(
                        gettext(
                            "There is more than one client signer on a challenge "
                            "transaction for an account that doesn't exist"
                        )
                    )
            except InvalidSep10ChallengeError as e:
                logger.info(
                    f"Missing or invalid signature(s) for {challenge.client_account_id}: {str(e)}"
                )
                return None, render_error_response(generic_err_msg % (str(e)))
            else:
                logger.info("Challenge verified using client's master key")
                return client_domain, None

        signers = account.load_ed25519_public_key_signers()
        threshold = account.thresholds.med_threshold
        try:
            signers_found = verify_challenge_transaction_threshold(
                challenge_transaction=envelope_xdr,
                server_account_id=settings.SIGNING_KEY,
                home_domains=settings.SEP10_HOME_DOMAINS,
                web_auth_domain=urlparse(settings.HOST_URL).netloc,
                network_passphrase=settings.STELLAR_NETWORK_PASSPHRASE,
                threshold=threshold,
                signers=signers,
            )
        except InvalidSep10ChallengeError as e:
            return None, render_error_response(generic_err_msg % (str(e)))

        logger.info(
            f"Challenge verified using account signers: {[s.account_id for s in signers_found]}"
        )
        return client_domain, None

    @staticmethod
    def _generate_jwt(envelope_xdr: str, client_domain: str = None) -> str:
        """
        Generates the JSON web token from the challenge transaction XDR.

        See: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0010.md#token
        """
        challenge = read_challenge_transaction(
            challenge_transaction=envelope_xdr,
            server_account_id=settings.SIGNING_KEY,
            home_domains=settings.SEP10_HOME_DOMAINS,
            web_auth_domain=urlparse(settings.HOST_URL).netloc,
            network_passphrase=settings.STELLAR_NETWORK_PASSPHRASE,
        )
        logger.info(
            f"Generating SEP-10 token for account {challenge.client_account_id}"
        )

        # set iat value to minimum timebound of the challenge so that the JWT returned
        # for a given challenge is always the same.
        # https://github.com/stellar/stellar-protocol/pull/982
        issued_at = challenge.transaction.transaction.preconditions.time_bounds.min_time

        # format sub value based on muxed account or memo
        if challenge.client_account_id.startswith("M") or not challenge.memo:
            sub = challenge.client_account_id
        else:
            sub = f"{challenge.client_account_id}:{challenge.memo}"

        jwt_dict = {
            "iss": os.path.join(settings.HOST_URL, "auth"),
            "sub": sub,
            "iat": issued_at,
            "exp": issued_at + 24 * 60 * 60,
            "jti": challenge.transaction.hash().hex(),
            "client_domain": client_domain,
        }
        return jwt.encode(jwt_dict, settings.SERVER_JWT_KEY, algorithm="HS256")

    @staticmethod
    def _get_client_signing_key(client_domain):
        client_toml_contents = fetch_stellar_toml(
            client_domain,
            client=RequestsClient(
                request_timeout=settings.SEP10_CLIENT_ATTRIBUTION_REQUEST_TIMEOUT
            ),
        )
        client_signing_key = client_toml_contents.get("SIGNING_KEY")
        if not client_signing_key:
            raise ValueError(gettext("SIGNING_KEY not present on 'client_domain' TOML"))
        try:
            Keypair.from_public_key(client_signing_key)
        except Ed25519PublicKeyInvalidError:
            raise ValueError(
                gettext("invalid SIGNING_KEY value on 'client_domain' TOML")
            )
        return client_signing_key



# Generated by Django 2.2.15 on 2020-08-14 17:31

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
import polaris.models
import uuid


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Asset",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created",
                    model_utils.fields.AutoCreatedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name="created",
                    ),
                ),
                (
                    "modified",
                    model_utils.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name="modified",
                    ),
                ),
                ("code", models.TextField()),
                (
                    "issuer",
                    models.TextField(
                        validators=[django.core.validators.MinLengthValidator(56)]
                    ),
                ),
                (
                    "significant_decimals",
                    models.IntegerField(
                        default=2,
                        validators=[
                            django.core.validators.MinValueValidator(0),
                            django.core.validators.MaxValueValidator(7),
                        ],
                    ),
                ),
                ("deposit_enabled", models.BooleanField(default=True)),
                (
                    "deposit_fee_fixed",
                    models.DecimalField(
                        blank=True, decimal_places=7, default=0, max_digits=30
                    ),
                ),
                (
                    "deposit_fee_percent",
                    models.DecimalField(
                        blank=True,
                        decimal_places=7,
                        default=0,
                        max_digits=30,
                        validators=[
                            django.core.validators.MinValueValidator(0),
                            django.core.validators.MaxValueValidator(100),
                        ],
                    ),
                ),
                (
                    "deposit_min_amount",
                    models.DecimalField(
                        blank=True, decimal_places=7, default=0, max_digits=30
                    ),
                ),
                (
                    "deposit_max_amount",
                    models.DecimalField(
                        blank=True,
                        decimal_places=7,
                        default=999999999999999999,
                        max_digits=30,
                    ),
                ),
                ("withdrawal_enabled", models.BooleanField(default=True)),
                (
                    "withdrawal_fee_fixed",
                    models.DecimalField(
                        blank=True, decimal_places=7, default=0, max_digits=30
                    ),
                ),
                (
                    "withdrawal_fee_percent",
                    models.DecimalField(
                        blank=True,
                        decimal_places=7,
                        default=0,
                        max_digits=30,
                        validators=[
                            django.core.validators.MinValueValidator(0),
                            django.core.validators.MaxValueValidator(100),
                        ],
                    ),
                ),
                (
                    "withdrawal_min_amount",
                    models.DecimalField(
                        blank=True, decimal_places=7, default=0, max_digits=30
                    ),
                ),
                (
                    "withdrawal_max_amount",
                    models.DecimalField(
                        blank=True,
                        decimal_places=7,
                        default=999999999999999999,
                        max_digits=30,
                    ),
                ),
                (
                    "send_fee_fixed",
                    models.DecimalField(
                        blank=True, decimal_places=7, default=0, max_digits=30
                    ),
                ),
                (
                    "send_fee_percent",
                    models.DecimalField(
                        blank=True,
                        decimal_places=7,
                        default=0,
                        max_digits=30,
                        validators=[
                            django.core.validators.MinValueValidator(0),
                            django.core.validators.MaxValueValidator(100),
                        ],
                    ),
                ),
                (
                    "send_min_amount",
                    models.DecimalField(
                        blank=True, decimal_places=7, default=0, max_digits=30
                    ),
                ),
                (
                    "send_max_amount",
                    models.DecimalField(
                        blank=True,
                        decimal_places=7,
                        default=999999999999999999,
                        max_digits=30,
                    ),
                ),
                ("distribution_seed", polaris.models.EncryptedTextField(null=True)),
                ("sep24_enabled", models.BooleanField(default=False)),
                ("sep6_enabled", models.BooleanField(default=False)),
                ("sep31_enabled", models.BooleanField(default=False)),
                ("symbol", models.TextField(default="$")),
            ],
        ),
        migrations.CreateModel(
            name="Transaction",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                ("paging_token", models.TextField(null=True)),
                (
                    "stellar_account",
                    models.TextField(
                        validators=[django.core.validators.MinLengthValidator(1)]
                    ),
                ),
                (
                    "kind",
                    models.CharField(
                        choices=[
                            ("deposit", "deposit"),
                            ("withdrawal", "withdrawal"),
                            ("send", "send"),
                        ],
                        default="deposit",
                        max_length=20,
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("pending_anchor", "pending_anchor"),
                            ("pending_trust", "pending_trust"),
                            ("pending_user", "pending_user"),
                            (
                                "pending_user_transfer_start",
                                "pending_user_transfer_start",
                            ),
                            ("incomplete", "incomplete"),
                            ("no_market", "no_market"),
                            ("too_small", "too_small"),
                            ("too_large", "too_large"),
                            ("pending_sender", "pending_sender"),
                            ("pending_receiver", "pending_receiver"),
                            (
                                "pending_transaction_info_update",
                                "pending_transaction_info_update",
                            ),
                            (
                                "pending_customer_info_update",
                                "pending_customer_info_update",
                            ),
                            ("completed", "completed"),
                            ("error", "error"),
                            ("pending_external", "pending_external"),
                            ("pending_stellar", "pending_stellar"),
                        ],
                        default="pending_external",
                        max_length=31,
                    ),
                ),
                ("status_eta", models.IntegerField(blank=True, null=True)),
                ("status_message", models.TextField(blank=True, null=True)),
                ("stellar_transaction_id", models.TextField(blank=True, null=True)),
                ("external_transaction_id", models.TextField(blank=True, null=True)),
                (
                    "amount_in",
                    models.DecimalField(
                        blank=True, decimal_places=7, max_digits=30, null=True
                    ),
                ),
                (
                    "amount_out",
                    models.DecimalField(
                        blank=True, decimal_places=7, max_digits=30, null=True
                    ),
                ),
                (
                    "amount_fee",
                    models.DecimalField(
                        blank=True, decimal_places=7, max_digits=30, null=True
                    ),
                ),
                ("started_at", models.DateTimeField(default=polaris.models.utc_now)),
                ("completed_at", models.DateTimeField(null=True)),
                ("from_address", models.TextField(blank=True, null=True)),
                ("to_address", models.TextField(blank=True, null=True)),
                ("required_info_updates", models.TextField(blank=True, null=True)),
                ("required_info_message", models.TextField(blank=True, null=True)),
                ("memo", models.TextField(blank=True, null=True)),
                (
                    "memo_type",
                    models.CharField(
                        choices=[("text", "text"), ("id", "id"), ("hash", "hash")],
                        default="text",
                        max_length=10,
                    ),
                ),
                ("receiving_anchor_account", models.TextField(blank=True, null=True)),
                ("refunded", models.BooleanField(default=False)),
                (
                    "protocol",
                    models.CharField(
                        choices=[
                            ("sep6", "sep6"),
                            ("sep24", "sep24"),
                            ("sep31", "sep31"),
                        ],
                        max_length=5,
                        null=True,
                    ),
                ),
                (
                    "asset",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="polaris.Asset"
                    ),
                ),
            ],
            options={
                "ordering": ("-started_at",),
            },
        ),
    ]



import sys
from typing import Callable

from polaris.integrations.customers import (
    CustomerIntegration,
    registered_customer_integration,
)
from polaris.integrations.fees import calculate_fee, registered_fee_func
from polaris.integrations.forms import TransactionForm, CreditCardForm
from polaris.integrations.info import default_info_func, registered_info_func
from polaris.integrations.quote import QuoteIntegration, registered_quote_integration
from polaris.integrations.rails import RailsIntegration, registered_rails_integration
from polaris.integrations.sep31 import (
    SEP31ReceiverIntegration,
    registered_sep31_receiver_integration,
)
from polaris.integrations.toml import get_stellar_toml, registered_toml_func
from polaris.integrations.transactions import (
    DepositIntegration,
    WithdrawalIntegration,
    registered_deposit_integration,
    registered_withdrawal_integration,
)
from polaris.integrations.custody import (
    CustodyIntegration,
    SelfCustodyIntegration,
    registered_custody_integration,
)


def register_integrations(
    deposit: DepositIntegration = None,
    withdrawal: WithdrawalIntegration = None,
    sep31_receiver: SEP31ReceiverIntegration = None,
    rails: RailsIntegration = None,
    toml: Callable = None,
    fee: Callable = None,
    sep6_info: Callable = None,
    customer: CustomerIntegration = None,
    custody: CustodyIntegration = None,
    quote: QuoteIntegration = None,
):
    """
    Registers the integration classes and functions with Polaris

    Call this function in your app's Django AppConfig.ready() function:
    ::

        from django.apps import AppConfig

        class PolarisIntegrationApp(AppConfig):
            name = 'Polaris Integration'
            verbose_name = name

            def ready(self):
                from polaris.integrations import register_integrations
                from myapp.integrations import (
                    MyDepositIntegration,
                    MyWithdrawalIntegration,
                    MyCustomerIntegration,
                    MyQuoteIntegration,
                    toml_integration,
                    fee_integrations,
                    info_integration
                )

                register_integrations(
                    deposit=MyDepositIntegration(),
                    withdrawal=MyWithdrawalIntegration(),
                    customer=MyCustomerIntegration(),
                    toml=toml_integration,
                    sep6_info=info_integration,
                    fee=fee_integration,
                    quote=MyQuoteIntegration()
                )

    Simply pass the integration classes or functions you use.

    :param deposit: the ``DepositIntegration`` subclass instance to be
        used by Polaris
    :param withdrawal: the ``WithdrawalIntegration`` subclass instance to
        be used by Polaris
    :param sep31_receiver: the ``SEP31ReceiverIntegration`` subclass instance
        to be used by Polaris
    :param rails: the ``RailsIntegration`` subclass instance to be used by
        Polaris
    :param toml: a function that returns stellar.toml data as a dictionary
    :param fee: a function that returns the fee that would be charged
    :param sep6_info: a function that returns the /info `fields` or `types`
        values for an Asset
    :param customer: the ``CustomerIntegration`` subclass instance to be used
        by Polaris
    :param custody: the ``CustodyIntegration`` subclass instance to be used
        by Polaris
    :param quote: the ``QuoteIntegration`` subclass instance to be used by
        Polaris
    :raises ValueError: missing argument(s)
    :raises TypeError: arguments are not subclasses of DepositIntegration or
        Withdrawal
    """
    this = sys.modules[__name__]

    if deposit and not issubclass(deposit.__class__, DepositIntegration):
        raise TypeError("deposit must be a subclass of DepositIntegration")
    elif withdrawal and not issubclass(withdrawal.__class__, WithdrawalIntegration):
        raise TypeError("withdrawal must be a subclass of WithdrawalIntegration")
    elif toml and not callable(toml):
        raise TypeError("toml parameter is not callable")
    elif fee and not callable(fee):
        raise TypeError("fee parameter is not callable")
    elif sep6_info and not callable(sep6_info):
        raise TypeError("info parameter is not callable")
    elif customer and not issubclass(customer.__class__, CustomerIntegration):
        raise TypeError("customer must be a subclass of CustomerIntegration")
    elif sep31_receiver and not issubclass(
        sep31_receiver.__class__, SEP31ReceiverIntegration
    ):
        raise TypeError("send must be a subclass of SEP31ReceiverIntegration")
    elif rails and not issubclass(rails.__class__, RailsIntegration):
        raise TypeError("rails must be a subclass of RailsIntegration")
    elif custody and not issubclass(custody.__class__, CustodyIntegration):
        raise TypeError("custody must be a subclass of CustodyIntegration")
    elif quote and not issubclass(quote.__class__, QuoteIntegration):
        raise TypeError("quote must be a subclass of QuoteIntegration")

    for obj, attr in [
        (deposit, "registered_deposit_integration"),
        (withdrawal, "registered_withdrawal_integration"),
        (toml, "registered_toml_func"),
        (fee, "registered_fee_func"),
        (sep6_info, "registered_info_func"),
        (customer, "registered_customer_integration"),
        (sep31_receiver, "registered_sep31_receiver_integration"),
        (rails, "registered_rails_integration"),
        (custody, "registered_custody_integration"),
        (quote, "registered_quote_integration"),
    ]:
        if obj:
            setattr(this, attr, obj)



from django.urls import re_path

from polaris.sep38 import info, prices, quote

urlpatterns = [
    re_path(r"^info/?$", info.info),
    re_path(r"^prices/?$", prices.get_prices),
    re_path(r"^price/?$", prices.get_price),
    re_path(r"^quote/?$", quote.post_quote),
    re_path(r"^quote/(?P<quote_id>[^/]+)/?$", quote.get_quote),
]



# Generated by Django 3.2.9 on 2022-02-11 06:24

from django.db import migrations, models
import polaris.models


class Migration(migrations.Migration):
    dependencies = [
        ("polaris", "0013_auto_20211011_1956"),
    ]

    operations = [
        migrations.CreateModel(
            name="PolarisHeartbeat",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("key", models.CharField(max_length=80, unique=True)),
                ("last_heartbeat", models.DateTimeField(blank=True, null=True)),
            ],
        ),
        migrations.AddField(
            model_name="transaction",
            name="queue",
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="transaction",
            name="queued_at",
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="transaction",
            name="submission_status",
            field=models.CharField(
                choices=[
                    ("not_ready", "not_ready"),
                    ("ready", "ready"),
                    ("processing", "processing"),
                    ("pending", "pending"),
                    ("pending_funding", "pending_funding"),
                    ("pending_trust", "pending_trust"),
                    ("blocked", "blocked"),
                    ("unblocked", "unblocked"),
                    ("completed", "completed"),
                    ("failed", "failed"),
                ],
                default="not_ready",
                max_length=31,
            ),
        ),
    ]



from rest_framework.decorators import api_view, renderer_classes
from rest_framework.renderers import JSONRenderer
from rest_framework.request import Request
from rest_framework.response import Response

from polaris.sep38.serializers import OffChainAssetSerializer
from polaris.models import Asset, OffChainAsset


@api_view(["GET"])
@renderer_classes([JSONRenderer])
def info(_: Request) -> Response:
    """
    Definition of the /info endpoint, in accordance with SEP-0038.
    See: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0038.md#response
    """
    info_data = {"assets": []}
    for asset in Asset.objects.filter(sep38_enabled=True):
        info_data["assets"].append({"asset": f"stellar:{asset.code}:{asset.issuer}"})
    info_data["assets"].extend(
        OffChainAssetSerializer(OffChainAsset.objects.all(), many=True).data
    )
    return Response(info_data)



from typing import Dict
from decimal import Decimal, DecimalException

from django.utils.translation import gettext
from rest_framework.decorators import api_view, renderer_classes
from rest_framework.renderers import JSONRenderer
from rest_framework.request import Request
from rest_framework.response import Response

from polaris.sep10.utils import validate_sep10_token
from polaris.sep10.token import SEP10Token
from polaris.integrations import registered_quote_integration as rqi
from polaris.utils import render_error_response, getLogger
from polaris.models import DeliveryMethod
from polaris.sep38.utils import (
    get_buy_assets,
    get_sell_asset,
    get_buy_asset,
    find_delivery_method,
)


logger = getLogger(__name__)


@api_view(["GET"])
@renderer_classes([JSONRenderer])
@validate_sep10_token()
def get_price(token: SEP10Token, request: Request) -> Response:
    try:
        request_data = validate_price_request(token, request)
    except ValueError as e:
        return render_error_response(str(e), status_code=400)

    try:
        price = rqi.get_price(**request_data)
    except ValueError as e:
        return render_error_response(str(e), status_code=400)
    except RuntimeError as e:
        return render_error_response(str(e), status_code=503)

    if not isinstance(price, Decimal):
        logger.error(
            "a non-Decimal price was returned from QuoteIntegration.get_price()"
        )
        return render_error_response(gettext("internal server error"), status_code=500)
    elif round(price, request_data["sell_asset"].significant_decimals) != price:
        logger.error(
            "the price returned from QuoteIntegration.get_price() did not have the correct "
            "number of significant decimals"
        )
        return render_error_response(gettext("internal server error"), status_code=500)

    if request_data.get("sell_amount"):
        sell_amount = request_data["sell_amount"]
        buy_amount = round(
            sell_amount / price, request_data["buy_asset"].significant_decimals
        )
    else:
        buy_amount = request_data["buy_amount"]
        sell_amount = round(
            price * buy_amount, request_data["sell_asset"].significant_decimals
        )

    return Response(
        {
            "price": str(round(price, request_data["sell_asset"].significant_decimals)),
            "sell_amount": str(sell_amount),
            "buy_amount": str(buy_amount),
        }
    )


@api_view(["GET"])
@renderer_classes([JSONRenderer])
@validate_sep10_token()
def get_prices(token: SEP10Token, request: Request) -> Response:
    try:
        request_data = validate_prices_request(token, request)
    except ValueError as e:
        return render_error_response(str(e), status_code=400)

    try:
        prices = rqi.get_prices(**request_data)
    except ValueError as e:
        return render_error_response(str(e), status_code=400)
    except RuntimeError as e:
        return render_error_response(str(e), status_code=503)

    if len(prices) != len(request_data["buy_assets"]):
        return render_error_response(gettext("internal server error"), status_code=500)

    buy_assets = []
    for idx, buy_asset in enumerate(request_data["buy_assets"]):
        price = prices[idx]
        buy_assets.append(
            {
                "asset": buy_asset.asset_identification_format,
                "price": str(
                    round(price, request_data["sell_asset"].significant_decimals)
                ),
                "decimals": request_data["sell_asset"].significant_decimals,
            }
        )

    return Response({"buy_assets": buy_assets})


def validate_prices_request(token: SEP10Token, request: Request) -> dict:
    validated_data: Dict = {"token": token, "request": request}
    required_fields = ["sell_asset", "sell_amount"]
    if not set(required_fields).issubset(request.GET.keys()):
        raise ValueError(
            gettext("missing required parameters. Required: ")
            + ", ".join(required_fields)
        )
    if request.GET.get("buy_delivery_method") and request.GET.get(
        "sell_delivery_method"
    ):
        raise ValueError(
            gettext(
                "'buy_delivery_method' or 'sell_delivery_method' "
                "is valid, but not both"
            )
        )
    validated_data["sell_asset"] = get_sell_asset(
        sell_asset_str=request.GET["sell_asset"],
        sell_delivery_method=request.GET.get("sell_delivery_method"),
        country_code=request.GET.get("country_code"),
    )
    validated_data["buy_assets"] = get_buy_assets(
        sell_asset=validated_data["sell_asset"],
        buy_delivery_method=request.GET.get("buy_delivery_method"),
        country_code=request.GET.get("country_code"),
    )
    if not validated_data["buy_assets"]:
        raise ValueError(
            gettext(
                "no 'buy_assets' for 'delivery_method' and 'country_code' specificed"
            )
        )
    try:
        validated_data["sell_amount"] = round(
            Decimal(request.GET["sell_amount"]),
            validated_data["sell_asset"].significant_decimals,
        )
    except DecimalException:
        raise ValueError(gettext("invalid 'sell_amount'; Expected decimal string."))
    # buy_assets will be empty if the following attributes are invalid
    # and an empty list will be returned prior to calling the anchor's
    # integration function.
    validated_data["buy_delivery_method"] = find_delivery_method(
        validated_data["buy_assets"][0],
        request.GET.get("buy_delivery_method"),
        DeliveryMethod.TYPE.buy,
    )
    validated_data["sell_delivery_method"] = find_delivery_method(
        validated_data["sell_asset"],
        request.GET.get("sell_delivery_method"),
        DeliveryMethod.TYPE.sell,
    )
    validated_data["country_code"] = request.GET.get("country_code")
    return validated_data


def validate_price_request(token: SEP10Token, request: Request) -> dict:
    validated_data = {"token": token, "request": request}
    required_fields = ["sell_asset", "buy_asset"]
    if not set(required_fields).issubset(request.GET.keys()):
        raise ValueError(
            gettext("missing required parameters. Required: ")
            + ", ".join(required_fields)
        )
    if not (bool(request.GET.get("buy_amount")) ^ bool(request.GET.get("sell_amount"))):
        raise ValueError(
            gettext("'sell_amount' or 'buy_amount' is required, but both is invalid")
        )
    if request.GET.get("buy_delivery_method") and request.GET.get(
        "sell_delivery_method"
    ):
        raise ValueError(
            gettext(
                "'buy_delivery_method' or 'sell_delivery_method' "
                "is valid, but not both"
            )
        )
    validated_data["sell_asset"] = get_sell_asset(
        sell_asset_str=request.GET["sell_asset"],
        sell_delivery_method=request.GET.get("sell_delivery_method"),
        country_code=request.GET.get("country_code"),
    )
    validated_data["buy_asset"] = get_buy_asset(
        sell_asset=validated_data["sell_asset"],
        buy_asset_str=request.GET["buy_asset"],
        buy_delivery_method=request.GET.get("buy_delivery_method"),
        country_code=request.GET.get("country_code"),
    )
    try:
        if request.GET.get("buy_amount"):
            validated_data["buy_amount"] = round(
                Decimal(request.GET["buy_amount"]),
                validated_data["buy_asset"].significant_decimals,
            )
        if request.GET.get("sell_amount"):
            validated_data["sell_amount"] = round(
                Decimal(request.GET["sell_amount"]),
                validated_data["sell_asset"].significant_decimals,
            )
    except DecimalException:
        raise ValueError(
            gettext("invalid 'buy_amount' or 'sell_amount'; Expected decimal strings.")
        )
    validated_data["country_code"] = request.GET.get("country_code")
    validated_data["buy_delivery_method"] = find_delivery_method(
        validated_data["buy_asset"],
        request.GET.get("buy_delivery_method"),
        DeliveryMethod.TYPE.buy,
    )
    validated_data["sell_delivery_method"] = find_delivery_method(
        validated_data["sell_asset"],
        request.GET.get("sell_delivery_method"),
        DeliveryMethod.TYPE.sell,
    )
    return validated_data



import signal
import time
import datetime
import asyncio
from decimal import Decimal
from enum import Enum
from typing import List, Optional, Dict, Union
from collections import defaultdict

import django.db.transaction
from django.core.management import BaseCommand
from django.db.models import Q
from stellar_sdk import (
    Keypair,
    ServerAsync,
    MuxedAccount,
    TransactionEnvelope,
    CreateAccount,
    CreateClaimableBalance,
)
from stellar_sdk.client.aiohttp_client import AiohttpClient
from stellar_sdk.exceptions import ConnectionError
from asgiref.sync import sync_to_async

from polaris import settings
from polaris.utils import (
    is_pending_trust,
    maybe_make_callback,
    maybe_make_callback_async,
    get_account_obj_async,
)
from polaris.integrations import (
    registered_deposit_integration as rdi,
    registered_rails_integration as rri,
    registered_custody_integration as rci,
    registered_fee_func,
    calculate_fee,
)

from polaris.exceptions import (
    TransactionSubmissionPending,
    TransactionSubmissionBlocked,
    TransactionSubmissionFailed,
)

from polaris.models import Transaction, PolarisHeartbeat
from polaris.utils import getLogger

logger = getLogger(__name__)


class TransactionType(Enum):
    DEPOSIT = 1
    CREATE_ACCOUNT = 2


SUBMIT_TRANSACTION_QUEUE = "SUBMIT_TRANSACTION_QUEUE"

DEFAULT_HEARTBEAT = 5
DEFAULT_INTERVAL = 10

RECOVER_LOCK_LOWER_BOUND = 30
PROCESS_PENDING_DEPOSITS_LOCK_KEY = "PROCESS_PENDING_DEPOSITS_LOCK"


class PolarisQueueAdapter:
    def __init__(self, queues):
        self.queues: Dict[str, asyncio.Queue] = {}
        for queue in queues:
            self.queues[queue] = asyncio.Queue()

    def populate_queues(self):
        """
        populate_queues gets called to read from the database and populate the in-memory queues
        """
        logger.debug("initializing queues from database...")
        ready_transactions = (
            Transaction.objects.filter(
                queue=SUBMIT_TRANSACTION_QUEUE,
                submission_status__in=[
                    Transaction.SUBMISSION_STATUS.ready,
                    Transaction.SUBMISSION_STATUS.processing,
                ],
                kind__in=[
                    Transaction.KIND.deposit,
                    getattr(Transaction.KIND, "deposit-exchange"),
                ],
                queued_at__isnull=False,
            )
            .order_by("queued_at")
            .select_related("asset")
        )

        logger.debug(
            f"found {len(ready_transactions)} transactions to queue for submit_transaction_task"
        )
        for transaction in ready_transactions:
            self.queue_transaction(
                "populate_queues", SUBMIT_TRANSACTION_QUEUE, transaction
            )

    def queue_transaction(self, source_task_name, queue_name, transaction):
        """
        Put the given transaction into a queue
        @param: source_task_name - the task that queued this transaction
        @param: queue_name - name of the queue to put the Transaction in
        @param: transaction - the Transaction to put in the queue
        """
        logger.debug(
            f"{source_task_name} - putting transaction {transaction.id} into {queue_name}"
        )
        self.queues[queue_name].put_nowait(transaction)

    async def get_transaction(self, source_task_name, queue_name) -> Transaction:
        """
        Consume a transaction from a queue
        @param: source_task_name - the task that is requesting a Transaction
        @param: queue_name - name of the queue to consume the Transaction from
        """
        logger.debug(f"{source_task_name} requesting task from queue: {queue_name}")
        transaction = await self.queues[queue_name].get()
        logger.debug(f"{source_task_name} got transaction: {transaction}")
        return transaction


class ProcessPendingDeposits:
    @classmethod
    async def check_rails_task(
        cls, queues: PolarisQueueAdapter, interval
    ):  # pragma: no cover
        """
        Periodically poll for deposit transactions that are ready to be processed
        and submit them to the CHECK_ACC_QUEUE for verification by the check_accounts_task.
        """
        logger.debug("check_rails_task started...")
        while True:
            await cls.check_rails_for_ready_transactions(queues)
            await asyncio.sleep(interval)

    @classmethod
    async def check_rails_for_ready_transactions(cls, queues: PolarisQueueAdapter):
        ready_transactions = await sync_to_async(cls.get_ready_deposits)()
        if not rci.account_creation_supported:
            Transaction.objects.filter(
                id__in=[t.id for t in ready_transactions]
            ).update(
                # TODO
                # we don't have an external status that indicates the user or wallet
                # needs to fund the account. Placing in pending_user for now.
                status=Transaction.STATUS.pending_user,
                submission_status=Transaction.SUBMISSION_STATUS.pending_funding,
            )
            return
        await cls.check_accounts(queues, ready_transactions)

    @classmethod
    async def check_accounts_task(
        cls, queues: PolarisQueueAdapter, interval: int
    ):  # pragma: no cover
        """
        Periodically polls accounts to determine if they exist on the Stellar
        Network. If they do, the transaction is queued for deposit submission,
        otherwise the transaction remains in the same state and is polled again
        at the provided interval.

        This task is only necessary if the registered CustodyIntegration class
        does not support account creation.
        """
        logger.debug("check_accounts_task started...")
        while True:
            transactions = await sync_to_async(cls.get_unfunded_account_transactions)()
            await cls.check_accounts(queues, transactions)
            await asyncio.sleep(interval)

    @classmethod
    async def check_accounts(
        cls, queues: PolarisQueueAdapter, transactions: List[Transaction]
    ):
        async with ServerAsync(settings.HORIZON_URI, client=AiohttpClient()) as server:
            for transaction in transactions:
                try:
                    _, account_json = await get_account_obj_async(
                        Keypair.from_public_key(transaction.to_address), server
                    )
                except RuntimeError:
                    # account not found, submitting the transaction will take care of account creation
                    await sync_to_async(cls.save_as_ready_for_submission)(transaction)
                    queues.queue_transaction(
                        "check_accounts_task", SUBMIT_TRANSACTION_QUEUE, transaction
                    )
                    continue
                except ConnectionError:
                    continue
                if (
                    not is_pending_trust(transaction, account_json)
                    or transaction.claimable_balance_supported
                ):
                    await sync_to_async(cls.save_as_ready_for_submission)(transaction)
                    queues.queue_transaction(
                        "check_accounts_task", SUBMIT_TRANSACTION_QUEUE, transaction
                    )
                else:
                    await sync_to_async(cls.save_as_pending_trust)(transaction)

    @staticmethod
    def get_unfunded_account_transactions():
        return list(
            Transaction.objects.filter(
                kind__in=[Transaction.KIND.deposit, "deposit-exchange"],
                submission_status=Transaction.SUBMISSION_STATUS.pending_funding,
            )
            .select_related("asset", "quote")
            .all()
        )

    @classmethod
    async def check_unblocked_transactions_task(
        cls, queues: PolarisQueueAdapter, interval: int
    ):
        """
        Get the transactions that are in a 'unblocked' submission_status and
        submit them to the SUBMIT_TRANSACTION_QUEUE for the submit_transactions_task to process.
        The 'unblocked' submission_status implies that Polaris preivously saved the
        transaction as 'blocked' due to a TransactionSubmissionBlocked exception being
        raised by a function that submits transactions to the Stellar Network.
        Anchors could manually resolve an issue causing the transaction to enter
        the 'blocked' status and update the transaction to be "unblocked", which would allow
        Polaris to detect and resubmit it.
        """
        logger.debug("check_unblocked_transactions_task started...")
        while True:
            await cls.process_unblocked_transactions(queues)
            await asyncio.sleep(interval)

    @classmethod
    async def process_unblocked_transactions(cls, queues: PolarisQueueAdapter):
        unblocked_transactions = await sync_to_async(cls.get_unblocked_transactions)()
        for transaction in unblocked_transactions:
            logger.info(
                f"check_unblocked_transactions_task - saving transaction {transaction.id} as 'ready'"
            )
            await sync_to_async(cls.save_as_ready_for_submission)(transaction)
            queues.queue_transaction(
                "check_unblocked_transactions_task",
                SUBMIT_TRANSACTION_QUEUE,
                transaction,
            )

    @staticmethod
    def save_as_ready_for_submission(transaction):
        logger.debug(f"saving transaction: {transaction.id} as 'ready'")
        transaction.queue = SUBMIT_TRANSACTION_QUEUE
        transaction.queued_at = datetime.datetime.now(datetime.timezone.utc)
        transaction.status = Transaction.STATUS.pending_anchor
        transaction.submission_status = Transaction.SUBMISSION_STATUS.ready
        transaction.save()

    @classmethod
    async def check_trustlines_task(
        cls, queues: PolarisQueueAdapter, interval: int
    ):  # pragma: no cover
        """
        For all transactions that are pending_trust, load the destination
        account json to determine if a trustline has been
        established. If a trustline for the requested asset is found, a the
        transaction is queued for submission.
        """
        logger.debug("check_trustlines_task started...")
        async with ServerAsync(settings.HORIZON_URI, client=AiohttpClient()) as server:
            while True:
                await cls.check_trustlines(queues, server)
                await asyncio.sleep(interval)

    @classmethod
    async def check_trustlines(cls, queues: PolarisQueueAdapter, server: ServerAsync):
        pending_trust_transactions: List[Transaction] = await sync_to_async(
            ProcessPendingDeposits.get_pending_trust_transactions
        )()
        for transaction in pending_trust_transactions:
            if transaction.to_address.startswith("M"):
                destination_account = MuxedAccount.from_account(
                    transaction.to_address
                ).account_id
            else:
                destination_account = transaction.to_address

            try:
                _, account_json = await get_account_obj_async(
                    Keypair.from_public_key(destination_account), server
                )
            except ConnectionError:
                logger.exception(f"failed to load account {destination_account}")
                continue

            if is_pending_trust(transaction=transaction, json_resp=account_json):
                continue

            logger.info(
                f"detected transaction {transaction.id} is no longer pending trust"
            )
            logger.debug(
                f"check_trustlines_task - saving transaction {transaction.id} as 'ready'"
            )
            if transaction.envelope_xdr:
                logger.info(
                    f"clearing submitted envelope_xdr for transaction {transaction.id}, "
                    f"envelope_xdr: {transaction.envelope_xdr}"
                )
                transaction.envelope_xdr = None
                transaction.stellar_transaction_id = None
            await sync_to_async(cls.save_as_ready_for_submission)(transaction)
            queues.queue_transaction(
                "check_trustlines_task", SUBMIT_TRANSACTION_QUEUE, transaction
            )

    @classmethod
    async def submit_transaction_task(
        cls, queues: PolarisQueueAdapter, locks: Dict
    ):  # pragma: no cover
        logger.debug("submit_transaction_task - running...")
        async with ServerAsync(settings.HORIZON_URI, client=AiohttpClient()) as server:
            while True:
                transaction = await queues.get_transaction(
                    "submit_transaction_task", SUBMIT_TRANSACTION_QUEUE
                )
                await cls.submit_transaction(transaction, server, locks, queues)

    @classmethod
    async def submit_transaction(
        cls,
        transaction: Transaction,
        server: ServerAsync,
        locks: Dict,
        queues: PolarisQueueAdapter,
    ):
        attempt = 1
        while True:
            logger.debug(
                f"submit_transaction_task calling submit() for transaction {transaction.id}, "
                f"attempt #{attempt}"
            )
            try:
                await ProcessPendingDeposits.submit(transaction, server, locks, queues)
            except TransactionSubmissionPending as e:
                await sync_to_async(cls.handle_submission_exception)(transaction, e)
                attempt += 1
                continue
            except (TransactionSubmissionBlocked, TransactionSubmissionFailed) as e:
                await sync_to_async(cls.handle_submission_exception)(transaction, e)
            except Exception as e:
                logger.exception("submit() threw an unexpected exception")
                message = getattr(e, "message", str(e))
                await sync_to_async(ProcessPendingDeposits.handle_error)(
                    transaction, f"{e.__class__.__name__}: {message}"
                )
                await maybe_make_callback_async(transaction)
            break

    @classmethod
    def get_ready_deposits(cls) -> List[Transaction]:
        """
        Polaris' API server processes deposit request and places the associated Transaction
        object in the `pending_user_transfer_start` status when all information necessary to
        submit the payment operation on Stellar has been collected.

        This function queries for these transaction, in addition to the Transaction objects
        that have been identified as pending external rails, and passes them to the
        DepositIntegration.poll_pending_deposits() integration function. Anchors return the
        transactions that are now available in their off-chain account and therefore ready
        for submission to the Stellar Network. Finally, this function performs various
        validations to ensure the transaction is truly ready and returns them.
        """
        pending_deposits = Transaction.objects.filter(
            status__in=[
                Transaction.STATUS.pending_user_transfer_start,
                Transaction.STATUS.pending_external,
            ],
            kind__in=[
                Transaction.KIND.deposit,
                getattr(Transaction.KIND, "deposit-exchange"),
            ],
        ).select_related("asset", "quote")

        ready_transactions = rri.poll_pending_deposits(pending_deposits)

        verified_ready_transactions = []
        for transaction in ready_transactions:
            if transaction.amount_fee is None or transaction.amount_out is None:
                if transaction.quote:
                    logger.error(
                        f"transaction {transaction.id} uses a quote but was returned "
                        "from poll_pending_deposits() without amount_fee or amount_out "
                        "assigned, skipping"
                    )
                    continue
                logger.warning(
                    f"transaction {transaction.id} was returned from "
                    f"poll_pending_deposits() without Transaction.amount_fee or "
                    f"Transaction.amount_out assigned. Future Polaris "
                    "releases will not calculate fees and delivered amounts"
                )

            asset = transaction.asset
            quote = transaction.quote
            transaction.refresh_from_db()
            transaction.asset = asset
            transaction.quote = quote
            if transaction.kind not in [
                transaction.KIND.deposit,
                getattr(transaction.KIND, "deposit-exchange"),
            ]:
                cls.handle_error(
                    transaction,
                    "poll_pending_deposits() returned a non-deposit transaction",
                )
                maybe_make_callback(transaction)
                continue
            if transaction.amount_in is None:
                cls.handle_error(
                    transaction,
                    "poll_pending_deposits() did not assign a value to the "
                    "amount_in field of a Transaction object returned",
                )
                maybe_make_callback(transaction)
                continue
            elif transaction.amount_fee is None:
                if registered_fee_func is calculate_fee:
                    try:
                        transaction.amount_fee = calculate_fee(
                            fee_params={
                                "amount": transaction.amount_in,
                                "operation": settings.OPERATION_DEPOSIT,
                                "asset_code": transaction.asset.code,
                            }
                        )
                    except ValueError:
                        transaction.amount_fee = Decimal(0)
                else:
                    transaction.amount_fee = Decimal(0)
                transaction.save()
            verified_ready_transactions.append(transaction)
        return verified_ready_transactions

    @staticmethod
    def get_pending_trust_transactions():
        """
        If the destination account does not have a trustline to the requested
        asset and the client application that initiated the request does not
        support claimable balances, Polaris places the transaction in the
        `pending_trust` status.

        The returned transactions will be submitted if their destination
        accounts now have a trustline to the asset.
        """
        return list(
            Transaction.objects.filter(
                kind__in=[Transaction.KIND.deposit, "deposit-exchange"],
                status=Transaction.STATUS.pending_trust,
                submission_status=Transaction.SUBMISSION_STATUS.pending_trust,
            ).select_related("asset", "quote")
        )

    @staticmethod
    def get_unblocked_transactions():
        """
        Return transactions that have been put in a SUBMISSION_STATUS.unblocked
        state.
        """
        unblocked = Q(submission_status=Transaction.SUBMISSION_STATUS.unblocked)
        got_signatures = Q(
            pending_signatures=False,
            envelope_xdr__isnull=False,
            status=Transaction.STATUS.pending_anchor,
        )
        unblocked_transactions = list(
            Transaction.objects.filter(
                unblocked | got_signatures,
                kind__in=[Transaction.KIND.deposit, "deposit-exchange"],
            )
            .select_related("asset", "quote")
            .exclude(
                submission_status__in=[
                    Transaction.SUBMISSION_STATUS.ready,
                    Transaction.SUBMISSION_STATUS.processing,
                ]
            )
        )
        for transaction in unblocked_transactions:
            logger.info(f"detected unblocked transaction: {transaction.id}")
        return unblocked_transactions

    @classmethod
    async def submit(
        cls,
        transaction: Transaction,
        server: ServerAsync,
        locks,
        queues: PolarisQueueAdapter,
    ):
        valid_statuses = [
            Transaction.STATUS.pending_user_transfer_start,
            Transaction.STATUS.pending_external,
            Transaction.STATUS.pending_anchor,
            Transaction.STATUS.pending_trust,
        ]
        if transaction.status not in valid_statuses:
            raise ValueError(
                f"Unexpected transaction status: {transaction.status}, expecting "
                f"{' or '.join(valid_statuses)}."
            )

        logger.info(f"initiating submission for {transaction.id}")
        transaction.status = Transaction.STATUS.pending_anchor
        transaction.submission_status = Transaction.SUBMISSION_STATUS.processing
        await sync_to_async(transaction.save)()
        await maybe_make_callback_async(transaction)

        try:
            distribution_account = await sync_to_async(rci.get_distribution_account)(
                asset=transaction.asset
            )
        except NotImplementedError:
            # Polaris has to assume that the custody service provider can handle concurrent
            # requests to send funds to destination accounts since it does not have a dedicated
            # distribution account.
            distribution_account = None
        else:
            # Aquire a lock for the source account of the transaction that will create the
            # deposit's destination account.
            logger.debug(
                f"requesting lock to submit deposit transaction {transaction.id}"
            )
            await locks["source_accounts"][distribution_account].acquire()
            logger.debug(
                f"locked to submit deposit transaction for transaction {transaction.id}"
            )

        try:
            try:
                _, destination_account_json = await get_account_obj_async(
                    Keypair.from_public_key(transaction.to_address), server
                )
            except RuntimeError:
                logger.info(
                    f"destination account: {transaction.to_address} not found, creating account..."
                )
                transaction_type = TransactionType.CREATE_ACCOUNT
                transaction_hash = await sync_to_async(rci.create_destination_account)(
                    transaction=transaction
                )
            else:
                has_trustline = not is_pending_trust(
                    transaction, destination_account_json
                )
                if not has_trustline and not transaction.claimable_balance_supported:
                    transaction.queue = None
                    transaction.queued_at = None
                    await sync_to_async(cls.save_as_pending_trust)(transaction)
                    if (
                        distribution_account in locks["source_accounts"]
                        and locks["source_accounts"][distribution_account].locked()
                    ):
                        logger.debug(
                            "unlocking after attempting submission for "
                            f"transaction {transaction.id}"
                        )
                        locks["source_accounts"][distribution_account].release()
                    return

                if transaction.envelope_xdr:
                    # If this is a multisig distribution account and there are two or more deposit
                    # transaction for the same destination account two "create_destination_account"
                    # transactions will be made for the same destination account. We already checked
                    # if the account exists so if we get to this part of the code and we see another
                    # create account operation, we clear the envelope_xdr and allow
                    # submit_deposit_transaction to generate a new transaction envelope.
                    signed_transaction = TransactionEnvelope.from_xdr(
                        transaction.envelope_xdr,
                        network_passphrase=settings.STELLAR_NETWORK_PASSPHRASE,
                    ).transaction
                    for op in signed_transaction.operations:
                        if isinstance(op, CreateAccount):
                            transaction.envelope_xdr = None
                            await sync_to_async(transaction.save)()

                transaction_type = TransactionType.DEPOSIT
                transaction_hash = await sync_to_async(rci.submit_deposit_transaction)(
                    transaction=transaction, has_trustline=has_trustline
                )
        finally:
            if (
                distribution_account in locks["source_accounts"]
                and locks["source_accounts"][distribution_account].locked()
            ):
                logger.debug(
                    "unlocking after attempting submission for "
                    f"transaction {transaction.id}"
                )
                locks["source_accounts"][distribution_account].release()

        transaction_json = (
            await server.transactions().transaction(transaction_hash).call()
        )

        if not transaction_json.get("successful"):
            await sync_to_async(cls.handle_error)(
                transaction,
                "transaction submission failed unexpectedly: "
                f"{transaction_json['result_xdr']}",
            )
            await maybe_make_callback_async(transaction)
        else:
            await cls.handle_successful_transaction(
                transaction_json=transaction_json,
                transaction=transaction,
                transaction_type=transaction_type,
                queues=queues,
            )

    @classmethod
    async def handle_successful_transaction(
        cls,
        transaction_type: TransactionType,
        transaction_json: dict,
        transaction: Transaction,
        queues: PolarisQueueAdapter,
    ):
        if transaction_type == TransactionType.DEPOSIT:
            await cls.handle_successful_deposit(
                transaction_json=transaction_json,
                transaction=transaction,
            )
        else:
            await cls.handle_successful_account_creation(
                transaction=transaction, queues=queues
            )

    @classmethod
    async def handle_successful_deposit(
        cls, transaction_json: dict, transaction: Transaction
    ):
        if transaction.claimable_balance_supported and rci.claimable_balances_supported:
            transaction.claimable_balance_id = cls.get_balance_id(transaction_json)

        transaction.paging_token = transaction_json["paging_token"]
        transaction.stellar_transaction_id = transaction_json["id"]
        transaction.status = Transaction.STATUS.completed
        transaction.submission_status = Transaction.SUBMISSION_STATUS.completed
        transaction.completed_at = datetime.datetime.now(datetime.timezone.utc)
        transaction.status_message = None
        transaction.queue = None
        transaction.queued_at = None
        if not transaction.quote:
            transaction.amount_out = round(
                Decimal(transaction.amount_in) - Decimal(transaction.amount_fee),
                transaction.asset.significant_decimals,
            )
        await sync_to_async(transaction.save)()
        logger.info(f"transaction {transaction.id} completed.")
        await maybe_make_callback_async(transaction)

        await sync_to_async(transaction.refresh_from_db)()
        try:
            await sync_to_async(rdi.after_deposit)(transaction=transaction)
        except NotImplementedError:
            pass
        except Exception:
            logger.exception("after_deposit() threw an unexpected exception")

        logger.info(f"deposit transaction: {transaction.id} successful")

    @classmethod
    async def handle_successful_account_creation(
        cls, transaction: Transaction, queues: PolarisQueueAdapter
    ):
        logger.info(
            f"account: {transaction.to_address} successfully created for transaction: {transaction.id}"
        )
        if transaction.claimable_balance_supported:
            await sync_to_async(cls.save_as_ready_for_submission)(transaction)
            queues.queue_transaction(
                "submit_transaction_task", SUBMIT_TRANSACTION_QUEUE, transaction
            )
        else:
            transaction.queue = None
            transaction.queued_at = None
            transaction.status_message = None
            await sync_to_async(cls.save_as_pending_trust)(transaction)

    @staticmethod
    def save_as_pending_trust(transaction: Transaction):
        logger.debug(f"saving transaction: {transaction.id} as 'pending_trust'")
        transaction.status = Transaction.STATUS.pending_trust
        transaction.submission_status = Transaction.SUBMISSION_STATUS.pending_trust
        transaction.save()

    @staticmethod
    def get_balance_id(response: dict) -> Optional[str]:
        """
        Pulls claimable balance ID from horizon responses if present

        The hex representation of the balanceID is important because it
        is the representation required to query and claim claimableBalances.

        :param
            response: the response from horizon

        :return
            hex representation of the balanceID or None
        """
        envelope = TransactionEnvelope.from_xdr(
            response["envelope_xdr"], settings.STELLAR_NETWORK_PASSPHRASE
        )
        balance_id = None
        for idx, op in enumerate(envelope.transaction.operations):
            if isinstance(op, CreateClaimableBalance):
                balance_id = envelope.transaction.get_claimable_balance_id(idx)
                break
        return balance_id

    @classmethod
    def handle_error(cls, transaction, message):
        transaction.queue = None
        transaction.queued_at = None
        transaction.submission_status = Transaction.SUBMISSION_STATUS.failed
        transaction.status_message = message
        transaction.status = Transaction.STATUS.error
        transaction.save()
        logger.error(f"transaction: {transaction.id} encountered an error: {message}")

    @classmethod
    def handle_submission_exception(cls, transaction, exception):
        if isinstance(exception, TransactionSubmissionBlocked):
            transaction.queue = None
            transaction.queued_at = None
            transaction.submission_status = Transaction.SUBMISSION_STATUS.blocked
            logger.info(f"transaction {transaction.id} is blocked, removing from queue")
        elif isinstance(exception, TransactionSubmissionFailed):
            transaction.queue = None
            transaction.queued_at = None
            transaction.status = Transaction.STATUS.error
            transaction.submission_status = Transaction.SUBMISSION_STATUS.failed
            logger.info(
                f"transaction {transaction.id} submission failed, "
                f"placing in error status"
            )
        elif isinstance(exception, TransactionSubmissionPending):
            transaction.submission_status = Transaction.SUBMISSION_STATUS.pending
            logger.info(f"transaction {transaction.id} is pending, resubmitting")
        transaction.status_message = str(exception)
        transaction.save()

    @classmethod
    def update_heartbeat(cls, key):
        PolarisHeartbeat.objects.filter(key=key).update(
            last_heartbeat=datetime.datetime.now(datetime.timezone.utc)
        )
        return

    @classmethod
    async def heartbeat_task(cls, key, heartbeat_interval):
        """
        Task that updates the given key at a specified interval (heatbeat_interval)
        """
        logger.debug("heartbeat_task started...")
        while True:
            await sync_to_async(ProcessPendingDeposits.update_heartbeat)(key)
            await asyncio.sleep(heartbeat_interval)

    @classmethod
    def acquire_lock(cls, key: str, heartbeat_interval: Union[int, float]):
        """
        This function creates a key in the database table 'polaris_polarisheartbeat'
        to ensure only one instance of the calling process runs at a given time. The key
        is deleted when the process exists gracefully. A 'heartbeat' is utilized in the event
        that the process does not exit gracefully. If the heartbeat's 'last_updated' value is
        5x longer than the typical heartbeat interval, it can be assumed that the previous
        process that created the key has crashed and a new lock can be acquired
        """
        attempt = 1
        while True:
            logger.debug(
                f"attempting to acquire lock on key: {key}, attempt #{attempt}..."
            )
            with django.db.transaction.atomic():
                heartbeat, created = PolarisHeartbeat.objects.get_or_create(key=key)
                if created:
                    # if the heartbeat key was created, update the last_heartbeat field to the current time
                    heartbeat.last_heartbeat = datetime.datetime.now(
                        datetime.timezone.utc
                    )
                    heartbeat.save()
                    logger.debug(
                        f"lock on key: {PROCESS_PENDING_DEPOSITS_LOCK_KEY} created"
                    )
                    return
                # the heartbeat key already exists (previous process did not shutdown gracefully), attempt
                # to acquire the lock based on time elapsed since the last heartbeat
                delta = (
                    datetime.datetime.now(datetime.timezone.utc)
                    - heartbeat.last_heartbeat
                )
                logger.debug(f"last heartbeat was {delta.total_seconds()} seconds ago")
                # the delta should be 5x the typical interval with a lower bound of 30 seconds
                if delta > max(
                    datetime.timedelta(seconds=heartbeat_interval * 5),
                    datetime.timedelta(seconds=RECOVER_LOCK_LOWER_BOUND),
                ):
                    heartbeat.last_heartbeat = datetime.datetime.now(
                        datetime.timezone.utc
                    )
                    heartbeat.save()
                    logger.debug(
                        f"lock on key: {PROCESS_PENDING_DEPOSITS_LOCK_KEY} acquired"
                    )
                    return
            logger.debug(
                f"unable to acquire lock on key: {key}, retrying in {heartbeat_interval} seconds..."
            )
            attempt += 1
            time.sleep(heartbeat_interval)

    @classmethod
    async def process_pending_deposits(  # pragma: no cover
        cls, task_interval: int, heartbeat_interval: int
    ):
        current_task = asyncio.current_task()
        signal.signal(
            signal.SIGINT,
            lambda signum, frame: asyncio.create_task(
                cls.exit_gracefully(signum, frame, current_task)
            ),
        )
        signal.signal(
            signal.SIGTERM,
            lambda signum, frame: asyncio.create_task(
                cls.exit_gracefully(signum, frame, current_task)
            ),
        )

        queues = PolarisQueueAdapter([SUBMIT_TRANSACTION_QUEUE])
        await sync_to_async(queues.populate_queues)()

        locks = {
            "source_accounts": defaultdict(asyncio.Lock),
            "destination_accounts": defaultdict(asyncio.Lock),
        }
        try:
            await asyncio.gather(
                ProcessPendingDeposits.heartbeat_task(
                    PROCESS_PENDING_DEPOSITS_LOCK_KEY, heartbeat_interval
                ),
                ProcessPendingDeposits.check_rails_task(queues, task_interval),
                ProcessPendingDeposits.check_accounts_task(queues, task_interval),
                ProcessPendingDeposits.check_trustlines_task(queues, task_interval),
                ProcessPendingDeposits.check_unblocked_transactions_task(
                    queues, task_interval
                ),
                ProcessPendingDeposits.submit_transaction_task(queues, locks),
            )
        except asyncio.CancelledError:
            logger.debug("caught root task CancelledError...")

    @classmethod
    async def exit_gracefully(cls, signal_name, frame, root_task):  # pragma: no cover
        logger.info(f"caught signal {signal_name}, cleaning up before exiting...")
        await sync_to_async(
            PolarisHeartbeat.objects.filter(
                key=PROCESS_PENDING_DEPOSITS_LOCK_KEY
            ).delete
        )()
        logger.debug(f"deleted heartbeat key: {PROCESS_PENDING_DEPOSITS_LOCK_KEY}...")
        current_task = asyncio.current_task()
        tasks = [
            task
            for task in asyncio.all_tasks()
            if task not in [current_task, root_task]
        ]
        for t in tasks:
            t.cancel()
        await asyncio.gather(*tasks, return_exceptions=True)
        logger.debug("all tasks have been canceled...")


class Command(BaseCommand):
    """
    This process handles all of the transaction submission logic for deposit transactions.

    When this command is invoked, Polaris queries the database for transactions in the
    following scenarios and processes them accordingly.

    A transaction is in the ``pending_user_transfer_start`` or ``pending_external`` status.
        Polaris passes these transaction the
        :meth:`~polaris.integrations.RailsIntegration.poll_pending_deposits` integration
        function, and the anchor is expected to return :class:`~polaris.models.Transaction`
        objects whose funds have been received off-chain. Polaris then checks if each
        transaction is in one of the secenarios outlined below, and if not, submits the
        return transactions them to the Stellar network. See the
        :meth:`~polaris.integrations.RailsIntegration.poll_pending_deposits()` integration
        function for more details.

    A transaction’s destination account does not have a trustline to the requested asset.
        Polaris checks if the trustline has been established. If it has, and the transaction’s
        source account doesn’t require multiple signatures, Polaris will submit the transaction
        to the Stellar Network.

    A transaction’s source account requires multiple signatures before submission to the network.
        In this case, :attr:`~polaris.models.Transaction.pending_signatures` is set to ``True``
        and the anchor is expected to collect signatures, save the transaction envelope to
        :attr:`~polaris.models.Transaction.envelope_xdr`, and set
        :attr:`~polaris.models.Transaction.pending_signatures` back to ``False``. Polaris will
        then query for these transactions and submit them to the Stellar network.

    **Optional arguments:**

        -h, --help            show this help message and exit
        --loop                Continually restart command after a specified number
                              of seconds.
        --interval INTERVAL, -i INTERVAL
                              The number of seconds to wait before restarting
                              command. Defaults to 10.
    """

    def add_arguments(self, parser):  # pragma: no cover
        parser.add_argument(
            "--interval",
            "-i",
            type=int,
            help="The number of seconds to wait before each internal periodic task executes."
            "Defaults to {}.".format(1),
        )

    def handle(self, *_args, **options):  # pragma: no cover
        """
        The entrypoint for the functionality implemented in this file.
        See diagram at polaris/docs/deployment
        """
        interval = options.get("interval") or DEFAULT_INTERVAL
        ProcessPendingDeposits.acquire_lock(
            PROCESS_PENDING_DEPOSITS_LOCK_KEY, DEFAULT_HEARTBEAT
        )
        asyncio.run(
            ProcessPendingDeposits.process_pending_deposits(interval, DEFAULT_HEARTBEAT)
        )
        logger.info("exiting after cleanup")



# Generated by Django 2.2.18 on 2021-02-26 23:53

from django.db import migrations, models
from django.db.models.functions import Concat
from polaris import settings
from polaris.shared.endpoints import SEP6_MORE_INFO_PATH
from polaris.sep24.urls import SEP24_MORE_INFO_PATH


def populate_more_info_urls(apps, _):
    Transaction = apps.get_model("polaris", "Transaction")
    if "sep-24" in settings.ACTIVE_SEPS:
        Transaction.objects.filter(protocol="sep24").update(
            more_info_url=Concat(
                models.Value(f"{settings.HOST_URL}{SEP24_MORE_INFO_PATH}?id="),
                models.F("id"),
                output_field=models.TextField(null=True, blank=True),
            ),
        )
    if "sep-6" in settings.ACTIVE_SEPS:
        Transaction.objects.filter(protocol="sep6").update(
            more_info_url=Concat(
                models.Value(f"{settings.HOST_URL}{SEP6_MORE_INFO_PATH}?id="),
                models.F("id"),
                output_field=models.TextField(null=True, blank=True),
            ),
        )


def noop(*_):
    # the DB column can be dropped without any reversal logic
    pass


class Migration(migrations.Migration):
    dependencies = [
        ("polaris", "0005_auto_20210226_0053"),
    ]

    operations = [
        migrations.AddField(
            model_name="transaction",
            name="more_info_url",
            field=models.TextField(blank=True, null=True),
        ),
        migrations.RunPython(populate_more_info_urls, noop),
    ]



import pytz
from datetime import datetime, timedelta, timezone

from rest_framework.decorators import api_view, parser_classes, renderer_classes
from rest_framework.parsers import JSONParser
from rest_framework.renderers import JSONRenderer
from rest_framework.request import Request
from rest_framework.response import Response
from django.contrib.sessions.backends.db import SessionStore

from polaris.utils import render_error_response, getLogger


logger = getLogger(__name__)


@api_view(["POST"])
@parser_classes([JSONParser])
@renderer_classes([JSONRenderer])
def post_tzinfo(request: Request) -> Response:
    if not (
        request.data.get("sessionId") and request.data.get("sessionOffset") is not None
    ):
        return render_error_response("missing required parameters")
    now = datetime.now(timezone.utc)
    offset = timedelta(minutes=request.data["sessionOffset"])
    zone = None
    for tz in map(pytz.timezone, pytz.all_timezones_set):
        if now.astimezone(tz).utcoffset() == offset:
            zone = tz.zone
            break
    if not zone:
        return render_error_response("no timezones matched with offset")
    session = SessionStore(session_key=request.data["sessionId"])
    session["timezone"] = zone
    session.save()
    return Response({"status": "ok", "tz": zone})



# Generated by Django 2.2.24 on 2021-08-17 20:37

from django.db import migrations, models


def match_amount_in(apps, _schema_editor):
    Transaction = apps.get_model("polaris", "Transaction")
    Transaction.objects.update(amount_expected=models.F("amount_in"))


def nop(*_):
    pass


class Migration(migrations.Migration):
    dependencies = [
        ("polaris", "0010_auto_20210809_1605"),
    ]

    operations = [
        migrations.AddField(
            model_name="transaction",
            name="amount_expected",
            field=models.DecimalField(
                blank=True, decimal_places=7, max_digits=30, null=True
            ),
        ),
        migrations.RunPython(match_amount_in, nop),
    ]



from django.contrib import admin

from polaris.models import (
    Transaction,
    Asset,
    Quote,
    OffChainAsset,
    DeliveryMethod,
    ExchangePair,
)


class TransactionAdmin(admin.ModelAdmin):
    """
    This defines the admin view of a Transaction.
    """

    list_display = "id", "asset_name", "kind", "status", "started_at"


class AssetAdmin(admin.ModelAdmin):
    """
    This defines the admin view of an Asset.
    """

    list_display = "code", "issuer", "deposit_enabled", "withdrawal_enabled"

    def get_fields(self, request, obj=None):
        fields = super().get_fields(request, obj)
        if not (
            request.user.is_superuser
            or request.user.user_permissions.filter(name="Can edit asset").exists()
        ):
            fields.remove("distribution_seed")
        return fields


admin.site.register(Transaction, TransactionAdmin)
admin.site.register(Asset, AssetAdmin)
admin.site.register(DeliveryMethod)
admin.site.register(Quote)
admin.site.register(OffChainAsset)
admin.site.register(ExchangePair)



from rest_framework import status
from rest_framework.request import Request
from rest_framework.response import Response

from polaris.sep10.token import SEP10Token
from polaris.utils import render_error_response


def check_auth(request, func, *args, **kwargs):
    """
    Check SEP 10 authentication in a request.
    Else call the original view function.
    """
    try:
        token = validate_jwt_request(request)
    except (ValueError, TypeError) as e:
        if "sep6/" in request.path:
            return Response({"type": "authentication_required"}, status=403)
        else:
            return render_error_response(str(e), status_code=status.HTTP_403_FORBIDDEN)
    return func(token, request, *args, **kwargs)


def validate_sep10_token():
    """Decorator to validate the SEP 10 token in a request."""

    def decorator(view):
        def wrapper(request, *args, **kwargs):
            return check_auth(request, view, *args, **kwargs)

        return wrapper

    return decorator


def validate_jwt_request(request: Request) -> SEP10Token:
    """
    Validate the JSON web token in a request and return the source account address

    :raises ValueError: invalid JWT
    """
    # While the SEP 24 spec calls the authorization header "Authorization", django middleware
    # renames this as "HTTP_AUTHORIZATION". We check this header for the JWT.
    jwt_header = request.headers.get("Authorization")
    if not jwt_header:
        raise ValueError("JWT must be passed as 'Authorization' header")
    bad_format_error = ValueError(
        "'Authorization' header must be formatted as 'Bearer <token>'"
    )
    if "Bearer" not in jwt_header:
        raise bad_format_error
    try:
        encoded_jwt = jwt_header.split(" ")[1]
    except IndexError:
        raise bad_format_error
    if not encoded_jwt:
        raise bad_format_error

    try:
        return SEP10Token(encoded_jwt)
    except ValueError as e:
        raise ValueError(f"SEP-10 token error: {str(e)}")



from django.urls import re_path
from django.views.decorators.csrf import csrf_exempt

from polaris.sep24.info import info
from polaris.sep24.fee import fee
from polaris.sep24.transaction import more_info, transaction, transactions
from polaris.sep24.withdraw import (
    withdraw,
    get_interactive_withdraw,
    post_interactive_withdraw,
    complete_interactive_withdraw,
)
from polaris.sep24.deposit import (
    deposit,
    complete_interactive_deposit,
    post_interactive_deposit,
    get_interactive_deposit,
)
from polaris.sep24.tzinfo import post_tzinfo

SEP24_MORE_INFO_PATH = "transaction/more_info"

urlpatterns = [
    re_path(r"^transactions/deposit/interactive/?$", csrf_exempt(deposit)),
    re_path(
        r"^transactions/deposit/interactive/complete/?$",
        complete_interactive_deposit,
        name="complete_interactive_withdraw",
    ),
    re_path(
        r"^transactions/deposit/webapp/submit/?$",
        post_interactive_deposit,
        name="post_interactive_deposit",
    ),
    re_path(
        r"^transactions/deposit/webapp/?$",
        get_interactive_deposit,
        name="get_interactive_deposit",
    ),
    re_path(r"^transactions/withdraw/interactive/?$", csrf_exempt(withdraw)),
    re_path(
        r"^transactions/withdraw/interactive/complete/?$",
        complete_interactive_withdraw,
        name="complete_interactive_withdraw",
    ),
    re_path(
        r"^transactions/withdraw/webapp/submit/?$",
        post_interactive_withdraw,
        name="post_interactive_withdraw",
    ),
    re_path(
        r"^transactions/withdraw/webapp/?$",
        get_interactive_withdraw,
        name="get_interactive_withdraw",
    ),
    re_path(r"^transactions/webapp/tzinfo/?$", post_tzinfo, name="tzinfo"),
    re_path(r"^info/?$", info),
    re_path(r"^fee/?$", fee),
    re_path(r"^transaction/?$", transaction),
    re_path(r"^transactions/?$", transactions),
    re_path(r"^" + SEP24_MORE_INFO_PATH + r"/?$", more_info, name="more_info"),
]



from typing import Dict, List, Optional
from decimal import Decimal

from rest_framework.request import Request

from polaris import settings
from polaris.models import Asset


def calculate_fee(
    fee_params: Dict, *_args: List, request: Optional[Request] = None, **_kwargs: Dict
) -> Decimal:
    """
    .. _`/fee`: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0024.md#fee

    Calculate the fee to be charged for the transaction described by `fee_params`.

    Note that this endpoint only supports calculating fees expressed in units of a
    Stellar asset. If fees cannot be calculated using the `fee_params` passed,
    raise a ``ValueError`` for Polaris to return a 400 Bad Request to the client.

    Replace this function with another by passing it to ``register_integrations()``
    as described in :ref:`api:Register Integrations`
    if the fees charged for transactions is not calculated using the asset's
    ``fee_fixed`` and ``fee_percent`` attributes.

    If replaced, `/info` responses will no longer contain the ``fee_fixed`` and
    ``fee_percent`` attributes per-asset. This is because Polaris can no longer
    assume fees are determined using those attributes alone.

    `fee_params` will always contain the following key-value pairs:

    - `amount`: ``Decimal``
    - `asset_code`: ``str``
    - `operation`: ``str``
    - `type`: ``str``

    Each of these key-value pairs correspond to the associated parameter for the
    `/fee`_ endpoint. The Decimal returned will be used as the `fee` value in the
    response.
    """
    amount = fee_params["amount"]
    asset = Asset.objects.filter(code=fee_params["asset_code"]).first()
    if fee_params["operation"] == settings.OPERATION_WITHDRAWAL:
        fee_percent = asset.withdrawal_fee_percent
        fee_fixed = asset.withdrawal_fee_fixed
    elif fee_params["operation"] == settings.OPERATION_DEPOSIT:
        fee_percent = asset.deposit_fee_percent
        fee_fixed = asset.deposit_fee_fixed
    elif fee_params["operation"] == settings.OPERATION_SEND:
        fee_percent = asset.send_fee_percent
        fee_fixed = asset.send_fee_fixed
    else:
        raise ValueError(f"unexpected operation: {fee_params['operation']}")

    if fee_fixed is None and fee_percent is None:
        return Decimal(0)

    if fee_fixed is None:
        fee_fixed = Decimal(0)
    elif fee_percent is None:
        fee_percent = Decimal(0)

    return round(
        fee_fixed + (fee_percent / Decimal("100") * Decimal(amount)),
        asset.significant_decimals,
    )


registered_fee_func = calculate_fee



"""This module defines custom management commands for the app admin."""
import asyncio
from asgiref.sync import sync_to_async
from typing import Dict, Optional, Union, List, Tuple
from decimal import Decimal

from django.core.management.base import BaseCommand
from django.db.models import Q
from stellar_sdk import FeeBumpTransactionEnvelope
from stellar_sdk.exceptions import NotFoundError
from stellar_sdk.transaction import Transaction as HorizonTransaction
from stellar_sdk.transaction_envelope import TransactionEnvelope
from stellar_sdk.utils import from_xdr_amount
from stellar_sdk.xdr import (
    PaymentResult,
    PathPaymentStrictSendResult,
    PathPaymentStrictReceiveResult,
    OperationResult,
    TransactionResult,
)
from stellar_sdk.operation import (
    Operation,
    Payment,
    PathPaymentStrictReceive,
    PathPaymentStrictSend,
)
from stellar_sdk.server_async import ServerAsync
from stellar_sdk.client.aiohttp_client import AiohttpClient

from polaris import settings
from polaris.models import Asset, Transaction
from polaris.utils import getLogger, maybe_make_callback_async
from polaris.integrations import registered_custody_integration as rci

logger = getLogger(__name__)
PaymentOpResult = Union[
    PaymentResult, PathPaymentStrictSendResult, PathPaymentStrictReceiveResult
]
PaymentOp = Union[Payment, PathPaymentStrictReceive, PathPaymentStrictSend]


class Command(BaseCommand):
    """
    Streams transactions to the :attr:`~polaris.models.Asset.distribution_account`
    of each :class:`~polaris.models.Asset` in the DB.

    Note that this command assumes Stellar payments are made to one distribution
    account address per asset. Some third party custody service providers may not
    use this scheme, in which case the custody integration class should provide
    an alternative command for detecting incoming Stellar payments.

    For every response from the server, attempts to find a matching transaction in
    the database and updates the transaction's status to ``pending_anchor`` or
    ``pending_receiver`` depending on the protocol.

    Then, the :mod:`~polaris.management.commands.execute_outgoing_transactions` process
    will query for transactions in those statuses and provide the anchor an integration
    function for executing the payment or withdrawal.

    **Optional arguments:**

        -h, --help            show this help message and exit
    """

    def handle(self, *_args, **_options):  # pragma: no cover
        try:
            asyncio.run(self.watch_transactions())
        except Exception as e:
            # This is very likely a bug, so re-raise the error and crash.
            # Heroku will restart the process unless it is repeatedly crashing,
            # in which case restarting isn't of much use.
            logger.exception("watch_transactions() threw an unexpected exception")
            raise e

    async def watch_transactions(self):  # pragma: no cover
        assets = await sync_to_async(list)(Asset.objects.all())
        await asyncio.gather(
            *[
                self._for_account(rci.get_distribution_account(asset=asset))
                for asset in assets
            ]
        )

    async def _for_account(self, account: str):
        """
        Stream transactions for the server Stellar address.
        """
        async with ServerAsync(settings.HORIZON_URI, client=AiohttpClient()) as server:
            try:
                # Ensure the distribution account actually exists
                await server.load_account(account)
            except NotFoundError:
                # This exception will crash the process, but the anchor needs
                # to provide valid accounts to watch.
                raise RuntimeError(
                    "Stellar distribution account does not exist in horizon"
                )

            last_completed_transaction = await sync_to_async(
                Transaction.objects.filter(
                    Q(kind=Transaction.KIND.withdrawal) | Q(kind=Transaction.KIND.send),
                    receiving_anchor_account=account,
                    status=Transaction.STATUS.completed,
                )
                .order_by("-completed_at")
                .first
            )()

            cursor = "0"
            if last_completed_transaction and last_completed_transaction.paging_token:
                cursor = last_completed_transaction.paging_token

            logger.info(
                f"starting transaction stream for {account} with cursor {cursor}"
            )
            endpoint = server.transactions().for_account(account).cursor(cursor)
            async for response in endpoint.stream():
                await self.process_response(response, account)

    @classmethod
    async def process_response(cls, response, account):
        # We should not match valid pending transactions with ones that were
        # unsuccessful on the stellar network. If they were unsuccessful, the
        # client is also aware of the failure and will likely attempt to
        # resubmit it, in which case we should match the resubmitted transaction
        if not response.get("successful"):
            return

        try:
            _ = response["id"]
            envelope_xdr = response["envelope_xdr"]
            memo = response["memo"]
            result_xdr = response["result_xdr"]
        except KeyError:
            return

        # Query filters for SEP6 and 24

        withdraw_filters = Q(
            status=Transaction.STATUS.pending_user_transfer_start,
            kind__in=[
                Transaction.KIND.withdrawal,
                getattr(Transaction.KIND, "withdrawal-exchange"),
            ],
        )
        # Query filters for SEP31
        send_filters = Q(
            status=Transaction.STATUS.pending_sender,
            kind=Transaction.KIND.send,
        )
        transactions = await sync_to_async(list)(
            Transaction.objects.filter(
                withdraw_filters | send_filters,
                memo=memo,
                receiving_anchor_account=account,
            )
            .select_related("asset")
            .all()
        )
        if not transactions:
            logger.info(f"No match found for stellar transaction {response['id']}")
            return
        elif len(transactions) == 1:
            transaction = transactions[0]
        else:
            # in the prior implementation of watch_transactions, the first transaction
            # to have the same memo is matched, so we'll do the same in the refactored
            # version.
            logger.error(f"multiple Transaction objects returned for memo: {memo}")
            transaction = transactions[0]

        logger.info(
            f"Matched transaction object {transaction.id} for stellar transaction {response['id']}"
        )

        try:
            horizon_tx = TransactionEnvelope.from_xdr(
                envelope_xdr,
                network_passphrase=settings.STELLAR_NETWORK_PASSPHRASE,
            ).transaction
            op_results = TransactionResult.from_xdr(result_xdr).result.results
        except ValueError:
            horizon_tx = FeeBumpTransactionEnvelope.from_xdr(
                envelope_xdr, network_passphrase=settings.STELLAR_NETWORK_PASSPHRASE
            ).transaction.inner_transaction_envelope.transaction
            op_results = TransactionResult.from_xdr(
                result_xdr
            ).result.inner_result_pair.result.result.results

        payment_data = await cls._find_matching_payment_data(
            response, horizon_tx, op_results, transaction
        )
        if not payment_data:
            logger.warning(f"Transaction matching memo {memo} has no payment operation")
            return

        # Transaction.amount_in is overwritten with the actual amount sent in the stellar
        # transaction. This allows anchors to validate the actual amount sent in
        # execute_outgoing_transactions() and handle invalid amounts appropriately.
        transaction.amount_in = round(
            Decimal(payment_data["amount"]),
            transaction.asset.significant_decimals,
        )

        # The stellar transaction has been matched with an existing record in the DB.
        # Now the anchor needs to initiate the off-chain transfer of the asset.
        if transaction.protocol == Transaction.PROTOCOL.sep31:
            # SEP-31 uses 'pending_receiver' status
            transaction.status = Transaction.STATUS.pending_receiver
            await sync_to_async(transaction.save)()
        else:
            # SEP-6 and 24 uses 'pending_anchor' status
            transaction.status = Transaction.STATUS.pending_anchor
            await sync_to_async(transaction.save)()
        await maybe_make_callback_async(transaction)
        return None

    @classmethod
    async def _find_matching_payment_data(
        cls,
        response: Dict,
        horizon_tx: HorizonTransaction,
        result_ops: List[OperationResult],
        transaction: Transaction,
    ) -> Optional[Dict]:
        matching_payment_data = None
        ops = horizon_tx.operations
        for idx, op_result in enumerate(result_ops):
            op, op_result = cls._cast_operation_and_result(ops[idx], op_result)
            if not op_result:  # not a payment op
                continue
            maybe_payment_data = cls._check_for_payment_match(
                op, op_result, transaction.asset, transaction
            )
            if maybe_payment_data:
                if ops[idx].source:
                    source = ops[idx].source.account_muxed or ops[idx].source.account_id
                else:
                    source = (
                        horizon_tx.source.account_muxed or horizon_tx.source.account_id
                    )
                await cls._update_transaction_info(
                    transaction, response["id"], response["paging_token"], source
                )
                matching_payment_data = maybe_payment_data
                break

        return matching_payment_data

    @classmethod
    async def _update_transaction_info(
        cls, transaction: Transaction, stellar_txid: str, paging_token: str, source: str
    ):
        transaction.stellar_transaction_id = stellar_txid
        transaction.from_address = source
        transaction.paging_token = paging_token
        await sync_to_async(transaction.save)()

    @classmethod
    def _check_for_payment_match(
        cls,
        operation: PaymentOp,
        op_result: PaymentOpResult,
        want_asset: Asset,
        transaction: Transaction,
    ) -> Optional[Dict]:
        payment_data = cls._get_payment_values(operation, op_result)
        if (
            payment_data["destination"] == transaction.receiving_anchor_account
            and payment_data["code"] == want_asset.code
            and payment_data["issuer"] == want_asset.issuer
        ):
            return payment_data
        else:
            return None

    @classmethod
    def _cast_operation_and_result(
        cls, operation: Operation, op_result: OperationResult
    ) -> Tuple[Optional[PaymentOp], Optional[PaymentOpResult]]:
        op_xdr_obj = operation.to_xdr_object()
        if isinstance(operation, Payment):
            return (
                Payment.from_xdr_object(op_xdr_obj),
                op_result.tr.payment_result,
            )
        elif isinstance(operation, PathPaymentStrictSend):
            return (
                PathPaymentStrictSend.from_xdr_object(op_xdr_obj),
                op_result.tr.path_payment_strict_send_result,
            )
        elif isinstance(operation, PathPaymentStrictReceive):
            return (
                PathPaymentStrictReceive.from_xdr_object(op_xdr_obj),
                op_result.tr.path_payment_strict_receive_result,
            )
        else:
            return None, None

    @classmethod
    def _get_payment_values(
        cls, operation: PaymentOp, op_result: PaymentOpResult
    ) -> Dict:
        values = {
            "destination": operation.destination.account_id,
            "amount": None,
            "code": None,
            "issuer": None,
        }
        if isinstance(operation, Payment):
            values["amount"] = str(operation.amount)
            values["code"] = operation.asset.code
            values["issuer"] = operation.asset.issuer
        elif isinstance(operation, PathPaymentStrictSend):
            # since the dest amount is not specified in a strict-send op,
            # we need to get the dest amount from the operation's result
            #
            # this method of fetching amounts gives the "raw" amount, so
            # we need to divide by Operation._ONE: 10000000
            # (Stellar uses 7 decimals places of precision)
            values["amount"] = from_xdr_amount(op_result.success.last.amount.int64)
            values["code"] = operation.dest_asset.code
            values["issuer"] = operation.dest_asset.issuer
        elif isinstance(operation, PathPaymentStrictReceive):
            values["amount"] = str(operation.dest_amount)
            values["code"] = operation.dest_asset.code
            values["issuer"] = operation.dest_asset.issuer
        else:
            raise ValueError("Unexpected operation, expected payment or path payment")
        return values



from django.urls import re_path
from polaris.sep12 import customer

urlpatterns = [
    re_path(r"^customer/callback/?$", customer.callback),
    re_path(r"^customer/verification/?$", customer.put_verification),
    re_path(r"^customer/(?P<account>[^/]+)/?$", customer.delete),
    re_path(r"^customer/?$", customer.CustomerAPIView.as_view()),
]



# Generated by Django 3.2.8 on 2021-10-11 19:56

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):
    dependencies = [
        ("polaris", "0012_auto_20210918_2053"),
    ]

    operations = [
        migrations.CreateModel(
            name="DeliveryMethod",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("type", models.TextField(choices=[("buy", "buy"), ("sell", "sell")])),
                ("name", models.TextField()),
                ("description", models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name="ExchangePair",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("buy_asset", models.TextField()),
                ("sell_asset", models.TextField()),
            ],
        ),
        migrations.AddField(
            model_name="asset",
            name="sep38_enabled",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="transaction",
            name="fee_asset",
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="asset",
            name="deposit_fee_fixed",
            field=models.DecimalField(
                blank=True, decimal_places=7, max_digits=30, null=True
            ),
        ),
        migrations.AlterField(
            model_name="asset",
            name="deposit_fee_percent",
            field=models.DecimalField(
                blank=True,
                decimal_places=7,
                max_digits=30,
                null=True,
                validators=[
                    django.core.validators.MinValueValidator(0),
                    django.core.validators.MaxValueValidator(100),
                ],
            ),
        ),
        migrations.AlterField(
            model_name="asset",
            name="send_fee_fixed",
            field=models.DecimalField(
                blank=True, decimal_places=7, max_digits=30, null=True
            ),
        ),
        migrations.AlterField(
            model_name="asset",
            name="send_fee_percent",
            field=models.DecimalField(
                blank=True,
                decimal_places=7,
                max_digits=30,
                null=True,
                validators=[
                    django.core.validators.MinValueValidator(0),
                    django.core.validators.MaxValueValidator(100),
                ],
            ),
        ),
        migrations.AlterField(
            model_name="asset",
            name="withdrawal_fee_fixed",
            field=models.DecimalField(
                blank=True, decimal_places=7, max_digits=30, null=True
            ),
        ),
        migrations.AlterField(
            model_name="asset",
            name="withdrawal_fee_percent",
            field=models.DecimalField(
                blank=True,
                decimal_places=7,
                max_digits=30,
                null=True,
                validators=[
                    django.core.validators.MinValueValidator(0),
                    django.core.validators.MaxValueValidator(100),
                ],
            ),
        ),
        migrations.AlterField(
            model_name="transaction",
            name="kind",
            field=models.CharField(
                choices=[
                    ("deposit", "deposit"),
                    ("withdrawal", "withdrawal"),
                    ("send", "send"),
                    ("deposit-exchange", "deposit-exchange"),
                    ("withdrawal-exchange", "withdrawal-exchange"),
                ],
                default="deposit",
                max_length=20,
            ),
        ),
        migrations.CreateModel(
            name="Quote",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4, primary_key=True, serialize=False
                    ),
                ),
                ("stellar_account", models.TextField()),
                ("account_memo", models.PositiveIntegerField(blank=True, null=True)),
                ("muxed_account", models.TextField(blank=True, null=True)),
                (
                    "type",
                    models.TextField(
                        choices=[("firm", "firm"), ("indicative", "indicative")]
                    ),
                ),
                ("sell_asset", models.TextField()),
                ("buy_asset", models.TextField()),
                ("sell_amount", models.DecimalField(decimal_places=7, max_digits=30)),
                (
                    "buy_amount",
                    models.DecimalField(
                        blank=True, decimal_places=7, max_digits=30, null=True
                    ),
                ),
                (
                    "price",
                    models.DecimalField(
                        blank=True, decimal_places=7, max_digits=30, null=True
                    ),
                ),
                ("expires_at", models.DateTimeField(blank=True, null=True)),
                ("country_code", models.TextField(blank=True, null=True)),
                ("requested_expire_after", models.DateTimeField(blank=True, null=True)),
                (
                    "buy_delivery_method",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="polaris.deliverymethod",
                    ),
                ),
                (
                    "sell_delivery_method",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="polaris.deliverymethod",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="OffChainAsset",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("scheme", models.TextField()),
                ("identifier", models.TextField()),
                ("significant_decimals", models.PositiveIntegerField(default=2)),
                ("country_codes", models.TextField(blank=True, null=True)),
                (
                    "delivery_methods",
                    models.ManyToManyField(to="polaris.DeliveryMethod"),
                ),
                ("symbol", models.TextField(blank=True, null=True)),
            ],
        ),
        migrations.AddConstraint(
            model_name="exchangepair",
            constraint=models.UniqueConstraint(
                fields=("buy_asset", "sell_asset"), name="exchangepair_unique_index"
            ),
        ),
        migrations.AddConstraint(
            model_name="deliverymethod",
            constraint=models.UniqueConstraint(
                fields=("name", "type"), name="deliverymethod_unique_index"
            ),
        ),
        migrations.AddField(
            model_name="transaction",
            name="quote",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="polaris.quote",
            ),
        ),
        migrations.AddConstraint(
            model_name="offchainasset",
            constraint=models.UniqueConstraint(
                fields=("scheme", "identifier"), name="offchain_unique_index"
            ),
        ),
    ]



from jwt import decode
from jwt.exceptions import InvalidTokenError
from datetime import datetime, timezone
from urllib.parse import urlparse
from typing import Union, Dict, Optional

from stellar_sdk import Keypair, MuxedAccount
from stellar_sdk.strkey import StrKey
from stellar_sdk.exceptions import (
    Ed25519PublicKeyInvalidError,
    MuxedEd25519AccountInvalidError,
)

from polaris import settings


class SEP10Token:
    """
    An object representing the authenticated session of the client.

    This object will be passed to every integration function that is called
    within the a request containing the JWT in the `Authorization` header.
    """

    _REQUIRED_FIELDS = {"iss", "sub", "iat", "exp"}

    def __init__(self, jwt: Union[str, Dict]):
        if isinstance(jwt, str):
            try:
                jwt = decode(jwt, settings.SERVER_JWT_KEY, algorithms=["HS256"])
            except InvalidTokenError as e:
                raise ValueError("unable to decode jwt" + str(e))
        elif not isinstance(jwt, Dict):
            raise ValueError(
                "invalid type for 'jwt' parameter: must be a string or dictionary"
            )

        if not self._REQUIRED_FIELDS.issubset(set(jwt.keys())):
            raise ValueError(
                f"jwt is missing one of the required fields: {', '.join(self._REQUIRED_FIELDS)}"
            )

        memo = None
        stellar_account = None
        if jwt["sub"].startswith("M"):
            try:
                StrKey.decode_muxed_account(jwt["sub"])
            except (MuxedEd25519AccountInvalidError, ValueError):
                raise ValueError(f"invalid muxed account address: {jwt['sub']}")
        elif ":" in jwt["sub"]:
            try:
                stellar_account, memo = jwt["sub"].split(":")
            except ValueError:
                raise ValueError(f"improperly formatted 'sub' value: {jwt['sub']}")
        else:
            stellar_account = jwt["sub"]

        if stellar_account:
            try:
                Keypair.from_public_key(stellar_account)
            except Ed25519PublicKeyInvalidError:
                raise ValueError(f"invalid Stellar public key: {jwt['sub']}")

        if memo:
            try:
                int(memo)
            except ValueError:
                raise ValueError(
                    f"invalid memo in 'sub' value, expected 64-bit integer: {memo}"
                )

        try:
            iat = datetime.fromtimestamp(jwt["iat"], tz=timezone.utc)
        except (OSError, ValueError, OverflowError):
            raise ValueError("invalid iat value")
        try:
            exp = datetime.fromtimestamp(jwt["exp"], tz=timezone.utc)
        except (OSError, ValueError, OverflowError):
            raise ValueError("invalid exp value")

        now = datetime.now(tz=timezone.utc)
        if now < iat or now > exp:
            raise ValueError("jwt is no longer valid")

        client_domain = jwt.get("client_domain")
        if (
            client_domain
            and urlparse(f"https://{client_domain}").netloc != client_domain
        ):
            raise ValueError("'client_domain' must be a hostname")

        self._payload = jwt

    @property
    def account(self) -> str:
        """
        The Stellar account (`G...`) authenticated. Note that a muxed account
        could have been authenticated, in which case `Token.muxed_account` should
        be used.
        """
        if self._payload["sub"].startswith("M"):
            return MuxedAccount.from_account(self._payload["sub"]).account_id
        elif ":" in self._payload["sub"]:
            return self._payload["sub"].split(":")[0]
        else:
            return self._payload["sub"]

    @property
    def muxed_account(self) -> Optional[str]:
        """
        The M-address specified in the payload's ``sub`` value, if present
        """
        return self._payload["sub"] if self._payload["sub"].startswith("M") else None

    @property
    def memo(self) -> Optional[int]:
        """
        The memo included with the payload's ``sub`` value, if present
        """
        return (
            int(self._payload["sub"].split(":")[1])
            if ":" in self._payload["sub"]
            else None
        )

    @property
    def issuer(self) -> str:
        """
        The principal that issued a token, RFC7519, Section 4.1.1 — a Uniform
        Resource Identifier (URI) for the issuer
        (https://example.com or https://example.com/G...)
        """
        return self._payload["iss"]

    @property
    def issued_at(self) -> datetime:
        """
        The time at which the JWT was issued RFC7519, Section 4.1.6 -
        represented as a UTC datetime object
        """
        return datetime.fromtimestamp(self._payload["iat"], tz=timezone.utc)

    @property
    def expires_at(self) -> datetime:
        """
        The expiration time on or after which the JWT will not accepted for
        processing, RFC7519, Section 4.1.4 — represented as a UTC datetime object
        """
        return datetime.fromtimestamp(self._payload["exp"], tz=timezone.utc)

    @property
    def client_domain(self) -> Optional[str]:
        """
        A nonstandard JWT claim containing the client's home domain, included if
        the challenge transaction contained a ``client_domain`` ManageData operation
        """
        return self._payload.get("client_domain")

    @property
    def payload(self) -> dict:
        """
        The decoded contents of the JWT string
        """
        return self._payload



from rest_framework.decorators import api_view, renderer_classes
from rest_framework.response import Response
from rest_framework.request import Request
from rest_framework.renderers import (
    TemplateHTMLRenderer,
    JSONRenderer,
    BrowsableAPIRenderer,
)

from django.views.decorators.clickjacking import xframe_options_exempt

from polaris.shared import endpoints
from polaris.sep10.utils import validate_sep10_token
from polaris.sep10.token import SEP10Token


@xframe_options_exempt
@api_view(["GET"])
@renderer_classes([TemplateHTMLRenderer])
def more_info(request: Request) -> Response:
    """
    Popup to display more information about a specific transaction.
    See table: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0024.md#4-customer-information-status
    """
    return endpoints.more_info(request)


@api_view(["GET"])
@renderer_classes([JSONRenderer, BrowsableAPIRenderer])
@validate_sep10_token()
def transactions(
    token: SEP10Token,
    request: Request,
) -> Response:
    """
    Definition of the /transactions endpoint, in accordance with SEP-0024.
    See: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0024.md#transaction-history
    """
    return endpoints.transactions_request(request, token)


@api_view(["GET"])
@renderer_classes([JSONRenderer, BrowsableAPIRenderer])
@validate_sep10_token()
def transaction(
    token: SEP10Token,
    request: Request,
) -> Response:
    """
    Definition of the /transaction endpoint, in accordance with SEP-0024.
    See: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0024.md#single-historical-transaction
    """
    return endpoints.transaction_request(request, token)



import uuid
from datetime import datetime, timezone
from decimal import Decimal, DecimalException
from typing import Dict, Union, Optional

from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.utils.translation import gettext
from rest_framework.parsers import JSONParser
from rest_framework.renderers import JSONRenderer
from rest_framework.decorators import parser_classes, renderer_classes, api_view
from rest_framework.request import Request
from rest_framework.response import Response

from polaris.models import Quote, Asset, OffChainAsset, DeliveryMethod
from polaris.integrations import registered_quote_integration as rqi
from polaris.sep10.token import SEP10Token
from polaris.sep10.utils import validate_sep10_token
from polaris.utils import render_error_response
from polaris.sep38.serializers import QuoteSerializer
from polaris.sep38.utils import (
    get_buy_asset,
    get_sell_asset,
    find_delivery_method,
)
from polaris.utils import getLogger
from polaris.settings import DATETIME_FORMAT


logger = getLogger(__name__)


@api_view(["GET"])
@renderer_classes([JSONRenderer])
@validate_sep10_token()
def get_quote(token: SEP10Token, request: Request, quote_id: str) -> Response:
    try:
        quote = Quote.objects.get(id=quote_id)
    except (ValidationError, ObjectDoesNotExist):
        return render_error_response("quote not found", status_code=404)
    return Response(QuoteSerializer(quote).data)


@api_view(["POST"])
@parser_classes([JSONParser])
@renderer_classes([JSONRenderer])
@validate_sep10_token()
def post_quote(token: SEP10Token, request: Request) -> Response:
    try:
        request_data = validate_quote_request(token, request)
    except ValueError as e:
        return render_error_response(str(e))

    quote = make_quote(**request_data)
    try:
        quote = rqi.post_quote(token=token, request=request, quote=quote)
    except ValueError as e:
        return render_error_response(str(e), status_code=400)
    except RuntimeError as e:
        return render_error_response(str(e), status_code=503)

    try:
        validate_quote_provided(
            quote,
            request.data.get("expire_after"),
            request_data["sell_asset"].significant_decimals,
        )
    except ValueError as e:
        logger.error(gettext("invalid quote provided: ") + str(e))
        return render_error_response("internal server error", status_code=500)

    if quote.sell_amount:
        quote.buy_amount = round(
            quote.sell_amount / quote.price,
            request_data["buy_asset"].significant_decimals,
        )
    else:
        quote.sell_amount = round(
            quote.price * quote.buy_amount,
            request_data["sell_asset"].significant_decimals,
        )

    quote.save()
    return Response(QuoteSerializer(quote).data, status=201)


def validate_quote_request(token: SEP10Token, request: Request) -> dict:
    validated_data: Dict = {"token": token, "request": request}
    required_fields = ["sell_asset", "buy_asset"]
    if not set(required_fields).issubset(request.data.keys()):
        raise ValueError(
            gettext("missing required parameters. Required: ")
            + ", ".join(required_fields)
        )
    if not (
        bool(request.data.get("buy_amount")) ^ bool(request.data.get("sell_amount"))
    ):
        raise ValueError(
            gettext("'sell_amount' or 'buy_amount' is required, but both is invalid")
        )
    if not (
        bool(request.data.get("buy_delivery_method"))
        ^ bool(request.data.get("sell_delivery_method"))
    ):
        raise ValueError(
            gettext(
                "'buy_delivery_method' or 'sell_delivery_method' "
                "is required, but both is invalid"
            )
        )
    if request.data.get("expire_after"):
        try:
            validated_data["expire_after"] = datetime.strptime(
                request.data.get("expire_after"), DATETIME_FORMAT
            ).replace(tzinfo=timezone.utc)
        except ValueError:
            raise ValueError(
                gettext(
                    "invalid 'expire_after' string format. "
                    "Expected UTC ISO 8601 datetime string."
                )
            )
        if validated_data["expire_after"] < datetime.now(timezone.utc):
            raise ValueError(
                gettext("invalid 'expire_after' datetime. Expected future datetime.")
            )
    else:
        validated_data["expire_after"] = None
    validated_data["sell_asset"] = get_sell_asset(
        sell_asset_str=request.data["sell_asset"],
        sell_delivery_method=request.data.get("sell_delivery_method"),
        country_code=request.data.get("country_code"),
    )
    validated_data["buy_asset"] = get_buy_asset(
        sell_asset=validated_data["sell_asset"],
        buy_asset_str=request.data["buy_asset"],
        buy_delivery_method=request.data.get("buy_delivery_method"),
        country_code=request.data.get("country_code"),
    )
    try:
        if request.data.get("buy_amount"):
            validated_data["buy_amount"] = round(
                Decimal(request.data["buy_amount"]),
                validated_data["buy_asset"].significant_decimals,
            )
        if request.data.get("sell_amount"):
            validated_data["sell_amount"] = round(
                Decimal(request.data["sell_amount"]),
                validated_data["sell_asset"].significant_decimals,
            )
    except DecimalException:
        raise ValueError(
            gettext("invalid 'buy_amount' or 'sell_amount'; Expected decimal strings.")
        )
    validated_data["country_code"] = request.data.get("country_code")
    validated_data["buy_delivery_method"] = find_delivery_method(
        validated_data["buy_asset"],
        request.data.get("buy_delivery_method"),
        DeliveryMethod.TYPE.buy,
    )
    validated_data["sell_delivery_method"] = find_delivery_method(
        validated_data["sell_asset"],
        request.data.get("sell_delivery_method"),
        DeliveryMethod.TYPE.sell,
    )
    return validated_data


def make_quote(
    token: SEP10Token,
    request: Request,
    sell_asset: Union[Asset, OffChainAsset],
    buy_asset: Union[Asset, OffChainAsset],
    buy_amount: Optional[Decimal] = None,
    sell_amount: Optional[Decimal] = None,
    sell_delivery_method: Optional[DeliveryMethod] = None,
    buy_delivery_method: Optional[DeliveryMethod] = None,
    country_code: Optional[str] = None,
    expire_after: Optional[datetime] = None,
):
    return Quote(
        id=str(uuid.uuid4()),
        stellar_account=token.account,
        account_memo=token.memo,
        muxed_account=token.muxed_account,
        type=Quote.TYPE.firm,
        sell_asset=sell_asset.asset_identification_format,
        buy_asset=buy_asset.asset_identification_format,
        sell_amount=sell_amount,
        buy_amount=buy_amount,
        buy_delivery_method=buy_delivery_method,
        sell_delivery_method=sell_delivery_method,
        country_code=country_code,
        requested_expire_after=expire_after,
    )


def validate_quote_provided(
    quote: Quote, requested_expire_after: str, sell_asset_significant_decimals: int
):
    if not isinstance(quote, Quote):
        raise ValueError("object returned is not a Quote")
    if quote.type != Quote.TYPE.firm:
        raise ValueError(f"quote is not of type '{Quote.TYPE.firm}'")
    if not quote.price:
        raise ValueError("quote must have price")
    if round(quote.price, sell_asset_significant_decimals) != quote.price:
        raise ValueError(
            "the price saved to Quote.price did not have the correct number "
            "of significant decimals"
        )
    if not (
        (quote.sell_amount is None or isinstance(quote.sell_amount, Decimal))
        and (quote.buy_amount is None or isinstance(quote.buy_amount, Decimal))
        and isinstance(quote.price, Decimal)
    ):
        raise ValueError("quote amounts must be of type decimal.Decimal")
    if not (bool(quote.sell_amount) ^ bool(quote.buy_amount)):
        raise ValueError(
            "polaris will calculate the amount not specified in the request"
        )
    if not (
        (quote.sell_amount is None or quote.sell_amount > 0)
        and (quote.buy_amount is None or quote.buy_amount > 0)
        and quote.price > 0
    ):
        raise ValueError("quote amounts must be positive")
    if not quote.expires_at:
        raise ValueError("quote must have expiration")
    elif quote.expires_at < datetime.now(timezone.utc):
        raise ValueError("quote expiration must be in the future")
    elif requested_expire_after and quote.expires_at < datetime.strptime(
        requested_expire_after, DATETIME_FORMAT
    ).replace(tzinfo=timezone.utc):
        raise ValueError("quote expiration must be at or after requested 'expire_at'")
    if not (bool(quote.buy_delivery_method) ^ bool(quote.sell_delivery_method)):
        raise ValueError(
            "quote must have either have buy_delivery_method or sell_delivery_method'"
        )
    if not (isinstance(quote.buy_asset, str) and isinstance(quote.sell_asset, str)):
        raise ValueError("quote assets must be strings")
    if not (
        quote.buy_asset.startswith("stellar") ^ quote.sell_asset.startswith("stellar")
    ):
        raise ValueError("quote must have one stellar asset and one off chain asset")



from typing import Union, Optional, List

from django.utils.translation import gettext
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Q

from polaris.models import OffChainAsset, Asset, ExchangePair, DeliveryMethod


def asset_id_to_kwargs(asset_id: str) -> dict:
    if asset_id.startswith("stellar"):
        _, code, issuer = asset_id.split(":")
        return {"code": code, "issuer": issuer}
    else:
        scheme, identifier = asset_id.split(":")
        return {"scheme": scheme, "identifier": identifier}


def is_stellar_asset(asset: str) -> bool:
    return asset.startswith("stellar")


def get_buy_assets(
    sell_asset: Union[Asset, OffChainAsset],
    buy_delivery_method: Optional[str],
    country_code: Optional[str],
) -> List[Union[Asset, OffChainAsset]]:
    asset_str = sell_asset.asset_identification_format
    pairs = ExchangePair.objects.filter(sell_asset=asset_str).all()
    if not pairs:
        return []
    buy_asset_strs = [p.buy_asset for p in pairs]
    conditions = Q()
    for asset_str in buy_asset_strs:
        conditions |= Q(**asset_id_to_kwargs(asset_str))
    if isinstance(sell_asset, Asset):
        kwargs = {}
        if country_code:
            kwargs["country_codes__icontains"] = country_code
        if buy_delivery_method:
            kwargs["delivery_methods__type"] = DeliveryMethod.TYPE.buy
            kwargs["delivery_methods__name"] = buy_delivery_method
        buy_assets = (
            OffChainAsset.objects.filter(conditions, **kwargs)
            .prefetch_related("delivery_methods")
            .all()
        )
    else:
        if buy_delivery_method:
            raise ValueError(
                gettext(
                    "unexpected 'buy_delivery_method', "
                    "client intends to buy a Stellar asset"
                )
            )
        buy_assets = Asset.objects.filter(conditions).all()
    return list(buy_assets)


def get_buy_asset(
    sell_asset: Union[Asset, OffChainAsset],
    buy_asset_str: str,
    buy_delivery_method: Optional[str],
    country_code: Optional[str],
) -> Union[Asset, OffChainAsset]:
    if isinstance(sell_asset, Asset):
        kwargs = {}
        try:
            scheme, identifier = buy_asset_str.split(":")
        except ValueError:
            raise ValueError(gettext("invalid 'buy_asset' format"))
        kwargs["scheme"] = scheme
        kwargs["identifier"] = identifier
        if country_code:
            kwargs["country_codes__icontains"] = country_code
        if buy_delivery_method:
            kwargs["delivery_methods__type"] = DeliveryMethod.TYPE.buy
            kwargs["delivery_methods__name"] = buy_delivery_method
        try:
            buy_asset = OffChainAsset.objects.prefetch_related("delivery_methods").get(
                **kwargs
            )
        except ObjectDoesNotExist:
            raise ValueError(
                gettext(
                    "unable to find 'buy_asset' using the following filters: "
                    "'country_code', 'buy_delivery_method'"
                )
            )
    else:
        if buy_delivery_method:
            raise ValueError(
                gettext(
                    "unexpected 'buy_delivery_method', "
                    "client intends to buy a Stellar asset"
                )
            )
        try:
            _, code, issuer = buy_asset_str.split(":")
        except ValueError:
            raise ValueError(gettext("invalid 'buy_asset' format"))
        try:
            buy_asset = Asset.objects.get(code=code, issuer=issuer)
        except ObjectDoesNotExist:
            raise ValueError(
                gettext(
                    "unable to find 'buy_asset' using the following filters: "
                    "'country_code', 'buy_delivery_method'"
                )
            )
    if not ExchangePair.objects.filter(
        sell_asset=sell_asset.asset_identification_format, buy_asset=buy_asset_str
    ).exists():
        raise ValueError(gettext("unsupported asset pair"))
    return buy_asset


def get_sell_asset(
    sell_asset_str: str,
    sell_delivery_method: Optional[str],
    country_code: Optional[str],
) -> Union[Asset, OffChainAsset]:
    try:
        if is_stellar_asset(sell_asset_str):
            if sell_delivery_method:
                raise ValueError(
                    gettext(
                        "unexpected 'sell_delivery_method', "
                        "client intends to sell a Stellar asset"
                    )
                )
            try:
                _, code, issuer = sell_asset_str.split(":")
            except ValueError:
                raise ValueError(gettext("invalid 'sell_asset' format"))
            return Asset.objects.get(code=code, issuer=issuer)
        else:
            try:
                scheme, identifier = sell_asset_str.split(":")
            except ValueError:
                raise ValueError(gettext("invalid 'sell_asset' format"))
            kwargs = {}
            if country_code:
                kwargs["country_codes__icontains"] = country_code
            if sell_delivery_method:
                kwargs["delivery_methods__type"] = DeliveryMethod.TYPE.sell
                kwargs["delivery_methods__name"] = sell_delivery_method
            return OffChainAsset.objects.prefetch_related("delivery_methods").get(
                scheme=scheme, identifier=identifier, **kwargs
            )
    except ObjectDoesNotExist:
        raise ValueError(
            gettext(
                "no 'sell_asset' for 'delivery_method' and 'country_code' specificed"
            )
        )


def find_delivery_method(
    asset: Union[Asset, OffChainAsset],
    delivery_method_name: str,
    delivery_method_type: str,
) -> Optional[DeliveryMethod]:
    if isinstance(asset, Asset):
        return None
    if not delivery_method_name:
        return None
    delivery_method = None
    for dm in asset.delivery_methods.all():
        if dm.type != delivery_method_type:
            continue
        if dm.name == delivery_method_name:
            delivery_method = dm
            break
    return delivery_method



from django.utils import translation
from django.utils.translation import get_supported_language_variant


def validate_or_use_default_language(lang: str) -> str:
    if not lang:
        return "en"
    try:
        get_supported_language_variant(lang)
    except LookupError:
        return "en"
    return lang


def activate_lang_for_request(lang: str):
    translation.activate(lang)



from django.utils.translation import gettext as _
from django.views.decorators.clickjacking import xframe_options_exempt
from django.core.exceptions import ValidationError, ObjectDoesNotExist
from rest_framework.decorators import api_view, renderer_classes
from rest_framework.response import Response
from rest_framework.request import Request
from rest_framework.renderers import (
    TemplateHTMLRenderer,
    JSONRenderer,
    BrowsableAPIRenderer,
)

from polaris.shared import endpoints
from polaris.sep10.utils import validate_sep10_token
from polaris.sep10.token import SEP10Token
from polaris.utils import (
    render_error_response,
    getLogger,
    validate_patch_request_fields,
)
from polaris.models import Transaction
from polaris.integrations import (
    registered_deposit_integration as rdi,
    registered_withdrawal_integration as rwi,
)

logger = getLogger(__name__)


@xframe_options_exempt
@api_view(["GET"])
@renderer_classes([TemplateHTMLRenderer])
def more_info(request: Request) -> Response:
    """
    Popup to display more information about a specific transaction.
    See table: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0024.md#4-customer-information-status
    """
    return endpoints.more_info(request, sep6=True)


@api_view(["GET"])
@renderer_classes([JSONRenderer, BrowsableAPIRenderer])
@validate_sep10_token()
def transactions(token: SEP10Token, request: Request) -> Response:
    """
    Definition of the /transactions endpoint, in accordance with SEP-0006.
    See: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0006.md#transaction-history
    """
    if token.account != request.GET.get("account"):
        return render_error_response(
            _("The account specified does not match authorization token"),
            status_code=403,
        )
    return endpoints.transactions_request(request, token, sep6=True)


@api_view(["GET"])
@renderer_classes([JSONRenderer, BrowsableAPIRenderer])
@validate_sep10_token()
def transaction(token: SEP10Token, request: Request) -> Response:
    """
    Definition of the /transaction endpoint, in accordance with SEP-0006.
    See: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0006.md#single-historical-transaction
    """
    return endpoints.transaction_request(request, token, sep6=True)


@api_view(["PATCH"])
@renderer_classes([JSONRenderer, BrowsableAPIRenderer])
@validate_sep10_token()
def patch_transaction(token: SEP10Token, request: Request, transaction_id: str):
    try:
        t = Transaction.objects.get(
            id=transaction_id,
            stellar_account=token.account,
            protocol=Transaction.PROTOCOL.sep6,
        )
    except (ValidationError, ObjectDoesNotExist):
        return render_error_response(_("transaction not found"), status_code=404)
    if t.status != Transaction.STATUS.pending_transaction_info_update:
        return render_error_response(_("update not required"))
    try:
        validate_patch_request_fields(request.data, t)
    except ValueError as e:
        return render_error_response(str(e))
    except RuntimeError as e:
        logger.exception(str(e))
        return render_error_response(_("unable to process request"), status_code=500)
    integration = rdi if t.kind == Transaction.KIND.deposit else rwi
    try:
        integration.patch_transaction(
            token=token, request=request, params=request.data, transaction=t
        )
    except ValueError as e:
        return render_error_response(str(e))
    except NotImplementedError:
        return render_error_response(_("not implemented"), status_code=501)
    t.status = Transaction.STATUS.pending_anchor
    t.required_info_updates = None
    t.required_info_message = None
    t.save()
    return Response(status=200)



# Generated by Django 2.2.18 on 2021-03-01 18:28

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("polaris", "0006_transaction_more_info_url"),
    ]

    operations = [
        migrations.AddField(
            model_name="transaction",
            name="on_change_callback",
            field=models.TextField(blank=True, null=True),
        ),
    ]



from typing import Dict

from rest_framework.request import Request

from polaris.utils import getLogger
from polaris.utils import SEP_9_FIELDS
from polaris.integrations import registered_customer_integration as rci
from polaris.models import Transaction
from polaris.sep10.token import SEP10Token
from polaris import settings


logger = getLogger(__name__)


def validate_403_response(
    token: SEP10Token,
    request: Request,
    integration_response: Dict,
    transaction: Transaction,
) -> Dict:
    """
    Ensures the response returned from `process_sep6_request()` matches the definitions
    described in SEP-6. This function can be used for both /deposit and /withdraw
    endpoints since the response schemas are identical.

    Note that this validation function is only for 403 responses. /deposit and /withdraw
    have distinct 200 Success response schemas so the validation for those are done in
    depost.py and withdraw.py.

    :param integration_response: the response dictionary returned from
        `process_sep6_request()`
    :param transaction: the transaction object that should not be saved to the DB
    :return: a new dictionary containing the valid key-value pairs from
        integration_response
    """
    if Transaction.objects.filter(id=transaction.id).exists():
        logger.error(
            "transaction cannot be saved when returning 403 SEP-6 deposit/withdraw response"
        )
        raise ValueError()

    statuses = ["pending", "denied"]
    types = ["customer_info_status", "non_interactive_customer_info_needed"]
    response = {"type": integration_response["type"]}
    if response["type"] not in types:
        logger.error("Invalid 'type' returned from process_sep6_request()")
        raise ValueError()

    elif response["type"] == types[0]:
        if integration_response.get("status") not in statuses:
            logger.error("Invalid 'status' returned from process_sep6_request()")
            raise ValueError()
        response["status"] = integration_response["status"]
        if settings.SEP6_USE_MORE_INFO_URL:
            more_info_url = rci.more_info_url(
                token=token,
                request=request,
                account=token.muxed_account or token.account,
                memo=token.memo,
            )
            response["more_info_url"] = more_info_url

    else:
        if "fields" not in integration_response:
            logger.error(f"missing 'fields' for {types[1]}")
            raise ValueError()
        elif not isinstance(integration_response["fields"], list):
            logger.error(f"invalid 'fields' for {types[1]}")
            raise ValueError()
        elif not all(f in SEP_9_FIELDS for f in integration_response["fields"]):
            logger.error(f"invalid 'fields' for {types[1]}")
            raise ValueError()
        else:
            response["fields"] = integration_response["fields"]

    return response



from datetime import timezone

from rest_framework import serializers

from polaris.models import Asset, Quote, OffChainAsset, DeliveryMethod
from polaris.settings import DATETIME_FORMAT


class QuoteSerializer(serializers.ModelSerializer):
    def __init__(self, data, *args, **kwargs):
        # cache each asset model fetched from the serializer methods
        self.sell_assets = {}
        self.buy_assets = {}
        super().__init__(data, *args, **kwargs)

    price = serializers.SerializerMethodField()
    sell_amount = serializers.SerializerMethodField()
    buy_amount = serializers.SerializerMethodField()
    expires_at = serializers.DateTimeField(
        format=DATETIME_FORMAT, default_timezone=timezone.utc
    )

    @staticmethod
    def cache_asset(asset: str, cache: dict):
        if asset not in cache:
            if asset.startswith("stellar"):
                _, code, issuer = asset.split(":")
                asset_obj = Asset.objects.get(code=code, issuer=issuer)
            else:
                scheme, identifier = asset.split(":")
                asset_obj = OffChainAsset.objects.get(
                    scheme=scheme, identifier=identifier
                )
            cache[asset] = asset_obj

    def get_price(self, instance):
        self.cache_asset(instance.sell_asset, self.sell_assets)
        return str(
            round(
                instance.price,
                self.sell_assets[instance.sell_asset].significant_decimals,
            )
        )

    def get_sell_amount(self, instance):
        self.cache_asset(instance.sell_asset, self.sell_assets)
        return str(
            round(
                instance.sell_amount,
                self.sell_assets[instance.sell_asset].significant_decimals,
            )
        )

    def get_buy_amount(self, instance):
        self.cache_asset(instance.buy_asset, self.buy_assets)
        return str(
            round(
                instance.buy_amount,
                self.buy_assets[instance.buy_asset].significant_decimals,
            )
        )

    class Meta:
        model = Quote
        fields = [
            "id",
            "price",
            "expires_at",
            "sell_asset",
            "buy_asset",
            "sell_amount",
            "buy_amount",
        ]


class OffChainAssetSerializer(serializers.ModelSerializer):
    asset = serializers.SerializerMethodField()
    country_codes = serializers.SerializerMethodField()
    sell_delivery_methods = serializers.SerializerMethodField()
    buy_delivery_methods = serializers.SerializerMethodField()

    @staticmethod
    def get_asset(instance):
        return instance.asset_identification_format

    @staticmethod
    def get_country_codes(instance):
        if instance.country_codes:
            return [cc.strip() for cc in instance.country_codes.split(",")]
        else:
            return []

    @staticmethod
    def get_sell_delivery_methods(instance):
        return DeliveryMethodSerializer(
            instance.delivery_methods.filter(type=DeliveryMethod.TYPE.sell),
            many=True,
        ).data

    @staticmethod
    def get_buy_delivery_methods(instance):
        return DeliveryMethodSerializer(
            instance.delivery_methods.filter(type=DeliveryMethod.TYPE.buy),
            many=True,
        ).data

    class Meta:
        model = OffChainAsset
        fields = [
            "asset",
            "country_codes",
            "sell_delivery_methods",
            "buy_delivery_methods",
        ]


class DeliveryMethodSerializer(serializers.ModelSerializer):
    class Meta:
        model = DeliveryMethod
        fields = ["name", "description"]



from decimal import Decimal, DecimalException
from urllib.parse import urlencode

from django.conf import settings as django_settings
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.core.validators import URLValidator
from django.urls import reverse
from django.views.decorators.clickjacking import xframe_options_exempt
from django.shortcuts import redirect
from django.utils.translation import gettext as _
from rest_framework.decorators import api_view, renderer_classes, parser_classes
from rest_framework.response import Response
from rest_framework.request import Request
from rest_framework.renderers import (
    TemplateHTMLRenderer,
    JSONRenderer,
    BrowsableAPIRenderer,
)
from rest_framework.parsers import MultiPartParser, FormParser, JSONParser
from stellar_sdk.keypair import Keypair
from stellar_sdk.strkey import StrKey
from stellar_sdk.exceptions import (
    Ed25519PublicKeyInvalidError,
    MuxedEd25519AccountInvalidError,
    ValueError as StellarSdkValueError,
)

from polaris import settings
from polaris.templates import Template
from polaris.utils import (
    getLogger,
    render_error_response,
    extract_sep9_fields,
    create_transaction_id,
    validate_account_and_memo,
)
from polaris.sep24.utils import (
    interactive_url,
    check_authentication,
    authenticate_session,
    invalidate_session,
    interactive_args_validation,
    get_timezone_utc_offset,
)
from polaris.sep10.utils import validate_sep10_token
from polaris.sep10.token import SEP10Token
from polaris.models import Asset, Transaction
from polaris.integrations.forms import TransactionForm
from polaris.locale.utils import (
    activate_lang_for_request,
    validate_or_use_default_language,
)
from polaris.integrations import (
    registered_withdrawal_integration as rwi,
    registered_custody_integration as rci,
    registered_fee_func,
    calculate_fee,
    registered_toml_func,
)

logger = getLogger(__name__)


@xframe_options_exempt
@api_view(["POST"])
@renderer_classes([TemplateHTMLRenderer])
@parser_classes([MultiPartParser, FormParser, JSONParser])
@check_authentication()
def post_interactive_withdraw(request: Request) -> Response:
    """
    POST /transactions/withdraw/webapp

    This endpoint processes form submissions during the withdraw interactive
    flow. The following steps are taken during this process:

        1. URL arguments are parsed and validated.
        2. form_for_transaction() is called to retrieve the form used to
           submit this request. This function is implemented by the anchor.
        3. The form is used to validate the data submitted, and if the form
           is a TransactionForm, the fee for the transaction is calculated.
        4. after_form_validation() is called to allow the anchor to process
           the data submitted. This function should change the application
           state such that the next call to form_for_transaction() returns
           the next form in the flow.
        5. form_for_transaction() is called again to retrieve the next
           form to be served to the user. If a form is returned, the
           function redirects to GET /transaction/withdraw/webapp. Otherwise,
           The user's session is invalidated, the transaction status is
           updated, and the function redirects to GET /more_info.
    """
    args_or_error = interactive_args_validation(request, Transaction.KIND.withdrawal)
    if "error" in args_or_error:
        return args_or_error["error"]

    transaction = args_or_error["transaction"]
    asset = args_or_error["asset"]
    callback = args_or_error["callback"]
    amount = args_or_error["amount"]

    form = rwi.form_for_transaction(
        request=request, transaction=transaction, post_data=request.data
    )
    if not form:
        logger.error(
            "Initial form_for_transaction() call returned None " f"for {transaction.id}"
        )
        if transaction.status != transaction.STATUS.incomplete:
            return render_error_response(
                _(
                    "The anchor did not provide content, is the interactive flow already complete?"
                ),
                status_code=422,
                as_html=True,
            )
        return render_error_response(
            _("The anchor did not provide content, unable to serve page."),
            status_code=500,
            as_html=True,
        )

    if not form.is_bound:
        # The anchor must initialize the form with request.data
        logger.error("form returned was not initialized with POST data, returning 500")
        return render_error_response(
            _("Unable to validate form submission."), status_code=500, as_html=True
        )

    elif form.is_valid():
        if issubclass(form.__class__, TransactionForm):
            transaction.amount_in = form.cleaned_data["amount"]
            transaction.amount_expected = form.cleaned_data["amount"]
            try:
                transaction.amount_fee = registered_fee_func(
                    request=request,
                    fee_params={
                        "amount": transaction.amount_in,
                        "type": form.cleaned_data.get("type"),
                        "operation": settings.OPERATION_WITHDRAWAL,
                        "asset_code": asset.code,
                    },
                )
            except ValueError:
                pass
            else:
                if settings.ADDITIVE_FEES_ENABLED:
                    transaction.amount_in += transaction.amount_fee
                    transaction.amount_expected += transaction.amount_fee
                transaction.amount_out = round(
                    transaction.amount_in - transaction.amount_fee,
                    asset.significant_decimals,
                )
            transaction.save()

        try:
            rwi.after_form_validation(
                request=request, form=form, transaction=transaction
            )
        except NotImplementedError:
            pass
        next_form = rwi.form_for_transaction(request=request, transaction=transaction)
        try:
            next_content = rwi.content_for_template(
                request=request,
                template=Template.WITHDRAW,
                form=next_form,
                transaction=transaction,
            )
        except NotImplementedError:
            next_content = None
        if next_form or next_content:
            args = {"transaction_id": transaction.id, "asset_code": asset.code}
            if amount:
                args["amount"] = amount
            if callback:
                args["callback"] = callback
            url = reverse("get_interactive_withdraw")
            return redirect(f"{url}?{urlencode(args)}")

        else:  # Last form has been submitted
            logger.info(
                f"Finished data collection and processing for transaction {transaction.id}"
            )
            invalidate_session(request)

            # ensure all changes from `after_form_validation()` have been saved to the db
            transaction.refresh_from_db()

            if transaction.status != transaction.STATUS.pending_anchor:
                # Add receiving account and memo now that anchor is ready to receive payment
                transaction.status = Transaction.STATUS.pending_user_transfer_start
                try:
                    receiving_account, memo, memo_type = validate_account_and_memo(
                        *rci.get_receiving_account_and_memo(
                            request=request, transaction=transaction
                        )
                    )
                except ValueError:
                    logger.exception(
                        "CustodyIntegration.get_receiving_account_and_memo() returned invalid values"
                    )
                    return render_error_response(
                        _("unable to process the request"), status_code=500
                    )
                transaction.receiving_anchor_account = receiving_account
                transaction.memo = memo
                transaction.memo_type = memo_type
                transaction.save()
            else:
                logger.info(f"Transaction {transaction.id} is pending KYC approval")

            args = {"id": transaction.id, "initialLoad": "true"}
            if callback:
                args["callback"] = callback
            return redirect(f"{reverse('more_info')}?{urlencode(args)}")

    else:
        try:
            content_from_anchor = (
                rwi.content_for_template(
                    request=request,
                    template=Template.WITHDRAW,
                    form=form,
                    transaction=transaction,
                )
                or {}
            )
        except NotImplementedError:
            content_from_anchor = {}

        url_args = {"transaction_id": transaction.id, "asset_code": asset.code}
        if callback:
            url_args["callback"] = callback
        if amount:
            url_args["amount"] = amount

        current_offset = get_timezone_utc_offset(
            request.session.get("timezone") or django_settings.TIME_ZONE
        )
        toml_data = registered_toml_func(request=request)
        post_url = f"{reverse('post_interactive_withdraw')}?{urlencode(url_args)}"
        content = {
            "form": form,
            "post_url": post_url,
            "operation": settings.OPERATION_WITHDRAWAL,
            "asset": asset,
            "symbol": asset.symbol,
            "show_fee_table": isinstance(form, TransactionForm),
            "use_fee_endpoint": registered_fee_func != calculate_fee,
            "org_logo_url": toml_data.get("DOCUMENTATION", {}).get("ORG_LOGO"),
            "additive_fees_enabled": settings.ADDITIVE_FEES_ENABLED,
            "timezone_endpoint": reverse("tzinfo"),
            "session_id": request.session.session_key,
            "current_offset": current_offset,
            **content_from_anchor,
        }
        return Response(
            content,
            template_name=content_from_anchor.get(
                "template_name", "polaris/withdraw.html"
            ),
            status=400,
        )


@api_view(["GET"])
@renderer_classes([TemplateHTMLRenderer])
@check_authentication()
def complete_interactive_withdraw(request: Request) -> Response:
    """
    GET /transactions/withdraw/interactive/complete

    This endpoint serves as a proxy to the
    ``WithdrawalIntegration.after_interactive_flow()`` function, which should be
    implemented if ``WithdrawalIntegration.interactive_url()`` is also implemented.

    Anchors using external interactive flows should redirect to this endpoint
    from the external application once complete so that the Polaris `Transaction`
    record can be updated with the appropriate information collected within the
    interactive flow.

    After allowing the anchor to process the information sent, this endpoint will
    redirect to the transaction's more info page, which will make the SEP-24
    callback request if requested by the client.

    Note that the more info page's template and related CSS should be updated
    if the anchor wants to keep the brand experience consistent with the external
    application seen by the user immediately prior.
    """
    transaction_id = request.query_params.get("transaction_id")
    callback = request.query_params.get("callback")
    try:
        transaction = Transaction.objects.get(id=transaction_id)
    except ObjectDoesNotExist:
        return render_error_response(
            _("transaction not found"), status_code=404, as_html=True
        )
    if callback and callback.lower() != "postmessage":
        try:
            URLValidator(schemes=["https", "http"])(callback)
        except ValidationError:
            return render_error_response(
                _("invalid 'callback' URL"), status_code=400, as_html=True
            )
    try:
        rwi.after_interactive_flow(request=request, transaction=transaction)
    except NotImplementedError:
        return render_error_response(_("internal error"), status_code=500, as_html=True)
    args = {"id": transaction_id, "initialLoad": "true"}
    if callback:
        args["callback"] = callback
    return redirect(f"{reverse('more_info')}?{urlencode(args)}")


@xframe_options_exempt
@api_view(["GET"])
@renderer_classes([TemplateHTMLRenderer])
@authenticate_session()
def get_interactive_withdraw(request: Request) -> Response:
    """
    GET /transactions/withdraw/webapp

    This endpoint retrieves the next form to be served to the user in the
    interactive flow. The following steps are taken during this process:

        1. URL arguments are parsed and validated.
        2. interactive_url() is called to determine whether or not the anchor
           uses an external service for the interactive flow. If a URL is
           returned, this function redirects to the URL. However, the session
           cookie should still be included in the response so future calls to
           GET /transactions/withdraw/interactive/complete are authenticated.
        3. form_for_transaction() is called to retrieve the next form to
           render to the user.
        4. get and post URLs are constructed with the appropriate arguments
           and passed to the response to be rendered to the user.
    """
    args_or_error = interactive_args_validation(request, Transaction.KIND.withdrawal)
    if "error" in args_or_error:
        return args_or_error["error"]

    transaction = args_or_error["transaction"]
    asset = args_or_error["asset"]
    callback = args_or_error["callback"]
    amount = args_or_error["amount"]
    lang = args_or_error["lang"]
    if args_or_error["on_change_callback"]:
        transaction.on_change_callback = args_or_error["on_change_callback"]
        transaction.save()

    try:
        url = rwi.interactive_url(
            request=request,
            transaction=transaction,
            asset=asset,
            amount=amount,
            callback=callback,
            lang=lang,
        )
    except NotImplementedError:
        pass
    else:
        # The anchor uses a standalone interactive flow
        return redirect(url)

    form = rwi.form_for_transaction(
        request=request, transaction=transaction, amount=amount
    )
    try:
        content_from_anchor = (
            rwi.content_for_template(
                request=request,
                template=Template.WITHDRAW,
                form=form,
                transaction=transaction,
            )
            or {}
        )
    except NotImplementedError:
        content_from_anchor = {}

    if not (form or content_from_anchor):
        logger.error("The anchor did not provide content, unable to serve page.")
        if transaction.status != transaction.STATUS.incomplete:
            return render_error_response(
                _(
                    "The anchor did not provide content, is the interactive flow already complete?"
                ),
                status_code=422,
                as_html=True,
            )
        return render_error_response(
            _("The anchor did not provide content, unable to serve page."),
            status_code=500,
            as_html=True,
        )

    url_args = {"transaction_id": transaction.id, "asset_code": asset.code}
    if callback:
        url_args["callback"] = callback
    if amount:
        url_args["amount"] = amount

    current_offset = get_timezone_utc_offset(
        request.session.get("timezone") or django_settings.TIME_ZONE
    )
    post_url = f"{reverse('post_interactive_withdraw')}?{urlencode(url_args)}"
    toml_data = registered_toml_func(request=request)
    content = {
        "form": form,
        "post_url": post_url,
        "operation": settings.OPERATION_WITHDRAWAL,
        "asset": asset,
        "symbol": asset.symbol,
        "show_fee_table": isinstance(form, TransactionForm),
        "use_fee_endpoint": registered_fee_func != calculate_fee,
        "org_logo_url": toml_data.get("DOCUMENTATION", {}).get("ORG_LOGO"),
        "additive_fees_enabled": settings.ADDITIVE_FEES_ENABLED,
        "timezone_endpoint": reverse("tzinfo"),
        "session_id": request.session.session_key,
        "current_offset": current_offset,
        **content_from_anchor,
    }

    response = Response(
        content,
        template_name=content_from_anchor.get("template_name", "polaris/withdraw.html"),
    )
    if lang:
        response.set_cookie(django_settings.LANGUAGE_COOKIE_NAME, lang)
    return response


@api_view(["POST"])
@validate_sep10_token()
@renderer_classes([JSONRenderer, BrowsableAPIRenderer])
@parser_classes([MultiPartParser, FormParser, JSONParser])
def withdraw(
    token: SEP10Token,
    request: Request,
) -> Response:
    """
    POST /transactions/withdraw/interactive

    Creates an `incomplete` withdraw Transaction object in the database and
    returns the URL entry-point for the interactive flow.
    """
    lang = validate_or_use_default_language(request.data.get("lang"))
    activate_lang_for_request(lang)

    asset_code = request.data.get("asset_code")
    source_account = request.data.get("account")
    sep9_fields = extract_sep9_fields(request.data)
    if not asset_code:
        return render_error_response(_("'asset_code' is required"))

    # Verify that the asset code exists in our database, with withdraw enabled.
    asset = Asset.objects.filter(code=asset_code).first()
    if not (asset and asset.withdrawal_enabled and asset.sep24_enabled):
        return render_error_response(_("invalid operation for asset %s") % asset_code)

    amount = None
    if request.data.get("amount"):
        try:
            amount = Decimal(request.data.get("amount"))
        except DecimalException:
            return render_error_response(_("invalid 'amount'"))
        if not (asset.withdrawal_min_amount <= amount <= asset.withdrawal_max_amount):
            return render_error_response(_("invalid 'amount'"))

    if source_account and source_account.startswith("M"):
        try:
            StrKey.decode_muxed_account(source_account)
        except (MuxedEd25519AccountInvalidError, StellarSdkValueError):
            return render_error_response(_("invalid 'account'"))
    elif source_account:
        try:
            Keypair.from_public_key(source_account)
        except Ed25519PublicKeyInvalidError:
            return render_error_response(_("invalid 'account'"))

    if sep9_fields:
        try:
            rwi.save_sep9_fields(
                token=token,
                request=request,
                stellar_account=token.account,
                muxed_account=token.muxed_account,
                account_memo=str(token.memo) if token.memo else None,
                account_memo_type=Transaction.MEMO_TYPES.id if token.memo else None,
                fields=sep9_fields,
                language_code=lang,
            )
        except ValueError as e:
            # The anchor found a validation error in the sep-9 fields POSTed by
            # the wallet. The error string returned should be in the language
            # specified in the request.
            return render_error_response(str(e))
        except NotImplementedError:
            # the KYC info passed via SEP-9 fields can be ignored if the anchor
            # wants to re-collect the information
            pass

    transaction_id = create_transaction_id()
    Transaction.objects.create(
        id=transaction_id,
        stellar_account=token.account,
        muxed_account=token.muxed_account,
        account_memo=token.memo,
        asset=asset,
        kind=Transaction.KIND.withdrawal,
        status=Transaction.STATUS.incomplete,
        memo_type=Transaction.MEMO_TYPES.hash,
        protocol=Transaction.PROTOCOL.sep24,
        more_info_url=request.build_absolute_uri(
            f"{reverse('more_info')}?id={transaction_id}"
        ),
        client_domain=token.client_domain,
        from_address=source_account,
    )
    logger.info(f"Created withdrawal transaction {transaction_id}")

    url = interactive_url(
        request=request,
        transaction_id=str(transaction_id),
        account=token.muxed_account or token.account,
        memo=token.memo,
        asset_code=asset_code,
        op_type=settings.OPERATION_WITHDRAWAL,
        amount=amount,
        lang=lang,
    )
    return Response(
        {"type": "interactive_customer_info_needed", "url": url, "id": transaction_id}
    )
